#include "config.h"

#include "libfspp.hpp"

#include "graphical_audio.hpp"

#include "search.hpp"
#include "graphics.hpp"

#include "touch.hpp"

using std::string;
using std::vector;
using std::list;

GraphicalAudio::GraphicalAudio()
  : db((conf->p_var_data_dir() + "audio.db").c_str()), imms_cur_path(""),
    extracted_all_files(false), exit_choose_cover_loop(false), reload_dirs(false)
{
  y_start = 70;

  image_width = 125;
  image_height = 96;

  if (conf->p_v_res() == 405) {
    image_width = 80;
    image_height = 67;
  }

  // text and space around image
  image_width_all = image_width+25;
  image_height_all = image_height+55;
  image_height_all_eks_text = image_height + 6;

  images_per_row = (conf->p_h_res()-2*25)/image_width_all;
  // lets make some more room for text
  image_width_all += ((conf->p_h_res()-2*25)%image_width_all)/images_per_row;

  rows = (conf->p_v_res()-2*y_start)/image_height_all;
  rows_search = (conf->p_v_res()-2*y_start-20)/image_height_all;

  // otherwise navigation won't work
  assert(rows > 1);
  assert(rows_search > 1);

  int to_spread_search = (conf->p_v_res()-2*y_start-20)-rows_search*image_height_all;
  image_height_all_search = to_spread_search/(rows_search+1) + image_height_all;

  int to_spread = (conf->p_v_res()-2*y_start)-rows*image_height_all;
  image_height_all += to_spread/(rows+1);
  y_start += to_spread/(rows+1);

  check_db();

#ifdef use_imms
  imms = new IMMSC();
  imms->start();
  if (!imms->check_connection())
    std::cerr << dgettext("mms-audio", "Failed to connect to IMMS") << std::endl;
  imms->setup(true);
#endif

  extracted_all_files = false;
}

GraphicalAudio::~GraphicalAudio()
{
#ifdef use_imms
  delete imms;
#endif

  db_mutex.enterMutex();
  db.close();
  db_mutex.leaveMutex();
}

void GraphicalAudio::create_db()
{
  vector<string> fields;
  fields.push_back("version INTEGER NOT NULL");
  db.createTable("Schema", fields);

  db.execute("INSERT INTO Schema VALUES('4')");

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("parent INTEGER");
  fields.push_back("filename TEXT");
  fields.push_back("name TEXT");
  fields.push_back("lname TEXT");
  fields.push_back("is_folder INTEGER"); // 1 = true
  db.createTable("Folders", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("name TEXT");
  fields.push_back("lname TEXT");
  db.createTable("Artist", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("name TEXT");
  fields.push_back("lname TEXT");
  db.createTable("Album", fields);

  // FIXME: use the next 3 rows for something

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("name INTEGER");
  db.createTable("Year", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("name TEXT");
  db.createTable("Genre", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("aid INTEGER NOT NULL");
  fields.push_back("gid INTEGER NOT NULL");
  db.createTable("GAudio", fields);

  fields.clear();
  fields.push_back("Path TEXT PRIMARY KEY");
  fields.push_back("Cover TEXT");
  db.createTable("Covers", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("Artist INTEGER");
  fields.push_back("Album INTEGER");
  fields.push_back("Title TEXT");
  fields.push_back("lTitle TEXT");
  fields.push_back("filename TEXT");
  fields.push_back("Bitrate FLOAT");
  fields.push_back("Length INTEGER");
  fields.push_back("Track INTEGER");
  db.createTable("Audio", fields);

  create_playlist_db();

  create_indexes();
}

void GraphicalAudio::create_playlist_db()
{
  vector<string> fields;
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("audio_id INTEGER");
  db.createTable("Playlist", fields);
}

void GraphicalAudio::create_indexes()
{
  db.execute("CREATE INDEX idx_album_name ON Album (lname)");

  db.execute("CREATE INDEX idx_artist_name ON Artist (lname)");

  db.execute("CREATE INDEX idx_folders_name ON Folders (name)");
  db.execute("CREATE INDEX idx_folders_lname ON Folders (lname)");
  db.execute("CREATE INDEX idx_folders_parent ON Folders (parent)");
  db.execute("CREATE INDEX idx_folders_filename ON Folders (filename)");

  db.execute("CREATE INDEX idx_audio_filename ON Audio (filename)");
  db.execute("CREATE INDEX idx_audio_title ON Audio (lTitle)");
  db.execute("CREATE INDEX idx_audio_artist ON Audio (Artist)");
  db.execute("CREATE INDEX idx_audio_album ON Audio (Album)");
}

void GraphicalAudio::check_db()
{
  if (db.hasTable("Schema")) {
    // check for correct version
    SQLQuery *q = db.query("Schema", "SELECT version FROM %t");
    if (q && q->numberOfTuples() > 0) {
      SQLRow &row = (*q)[0];
      if (row["version"] != "4") {

	Print print(gettext("Upgrading audio database schema"), Print::INFO);

	SQLQuery *dq = db.query("sqlite_master", "SELECT name FROM %t WHERE type='table'");
	if (dq) {
	  for (int i = 0; i < dq->numberOfTuples(); ++i) {
	    db.execute(("DROP TABLE " + (*dq)[i]["name"]).c_str());
	  }
	}

	delete dq;

	dq = db.query("sqlite_master", "SELECT name FROM %t WHERE type='index'");
	if (dq) {
	  for (int i = 0; i < dq->numberOfTuples(); ++i) {
	    db.execute(("DROP INDEX " + (*dq)[i]["name"]).c_str());
	  }
	}

	delete dq;

	create_db();
      }
    }
    delete q;
  } else {
    // create the whole db
    create_db();
  }
}

void GraphicalAudio::reparse_file(const string& path)
{
  db_mutex.enterMutex();

  char *tmp = sqlite3_mprintf("DELETE FROM Folders WHERE filename LIKE '%q%%'", path.c_str());
  db.execute(tmp);
  sqlite3_free(tmp);

  tmp = sqlite3_mprintf("DELETE FROM Audio WHERE filename LIKE '%q%%'", path.c_str());
  db.execute(tmp);
  sqlite3_free(tmp);

  db.execute("DELETE FROM Album WHERE id NOT IN (SELECT DISTINCT Album FROM Audio)");
  db.execute("DELETE FROM Artist WHERE id NOT IN (SELECT DISTINCT Artist FROM Audio)");

  db_mutex.leaveMutex();

  // reload metadata
  Dbaudiofile die(path);
}

bool GraphicalAudio::reload_dir(const string& path)
{
  // generate new list
  vector<Dbaudiofile> hdfiles = rdir_hd(path, false);

  vector<Dbaudiofile> hd_sorted_by_id;
  add_tracks_from_dir(hdfiles, hd_sorted_by_id);

  vector<Dbaudiofile> dbfiles = rdir(path);

  vector<Dbaudiofile> db_sorted_by_id;
  add_tracks_from_dir(dbfiles, db_sorted_by_id);

  // find what is new and what is removed
  vector<string> new_list;
  foreach (Dbaudiofile& file, hd_sorted_by_id)
    new_list.push_back(file.path);

  vector<string> old_list;
  foreach (Dbaudiofile& file, db_sorted_by_id)
    old_list.push_back(file.path);

  vector<string> new_files;
  vector<string> removed_files;

  folder_difference(new_list, old_list, new_files, removed_files);

  // insert new into db
  foreach (string& path, new_files) {

    if (path[path.size() - 1] == '/')
      path = path.substr(0, path.size()-1);

    string parent_path = "";
    string::size_type pos_backslash = path.rfind('/');
    if (pos_backslash != string::npos) {
      parent_path = path.substr(0, pos_backslash+1);
    }

    insert_file_into_db(path, parent_path);
  }

  // remove old from db
  foreach (string& path, removed_files) {

    db_mutex.enterMutex();

    char *tmp = sqlite3_mprintf("DELETE FROM Folders WHERE filename LIKE '%q%%'", path.c_str());
    db.execute(tmp);
    sqlite3_free(tmp);

    tmp = sqlite3_mprintf("DELETE FROM Audio WHERE filename LIKE '%q%%'", path.c_str());
    db.execute(tmp);
    sqlite3_free(tmp);

    db_mutex.leaveMutex();

    // and from playlist
    for (vector<Dbaudiofile>::iterator pi = playlist.begin(); pi != playlist.end();) {
      if (pi->path.find(path) != string::npos) {
	remove_track_from_playlist(*pi);
	pi = playlist.erase(pi);
      } else
	++pi;
    }
  }

  if (removed_files.size() > 0) {
    // sync the other db's
    db_mutex.enterMutex();
    db.execute("DELETE FROM Album WHERE id NOT IN (SELECT DISTINCT Album FROM Audio)");
    db.execute("DELETE FROM Artist WHERE id NOT IN (SELECT DISTINCT Artist FROM Audio)");
    db_mutex.leaveMutex();
  }

  return (removed_files.size() > 0 || new_files.size() > 0);
}

void GraphicalAudio::reload_current_dirs()
{
  if (mode == PLAY && playlist.size() == 0) {
    mode = ADD;
    files = &audio_list;
  }
  else if (playlist_pos_int > playlist.size()-1)
    playlist_pos_int = playlist.size()-1;

  load_current_dirs();

  if (audio_list.size() == 0)
    one_up();
}

string GraphicalAudio::id_to_filename(int db_id)
{
  string path;

  db_mutex.enterMutex();

  SQLQuery *q = db.query("Folders", ("SELECT filename FROM %t WHERE id='" + conv::itos(db_id) + "'").c_str());
  if (q && q->numberOfTuples() > 0)
    path = (*q)[0]["filename"];
  else
    assert(false);

  db_mutex.leaveMutex();

  return path;
}

bool GraphicalAudio::change_dir_to_id(int db_id)
{
  db_mutex.enterMutex();

  SQLQuery *q = db.query("Folders", ("SELECT filename FROM %t WHERE id='" + conv::itos(db_id) + "'").c_str());
  if (q && q->numberOfTuples() > 0) {

    list<string> old_watch_folder = folders.top().first;

    SQLRow &row = (*q)[0];

    string filename = row["filename"];
    string cut_filename = filename;
    string name_in_folder, path, current_path;

    delete q;
    db_mutex.leaveMutex();

    if (filename[filename.size()-1] == '/')
      cut_filename = filename.substr(0, filename.size()-1);

    // FIXME: restore?
    for (int i = 0, size = folders.size(); i < size; ++i)
      folders.pop();

    string::size_type pos_backslash = cut_filename.rfind('/');
    if (pos_backslash != string::npos) {
      name_in_folder = cut_filename.substr(pos_backslash+1);
      path = cut_filename.substr(0, pos_backslash+1);
    }

    // FIXME: restore?
    audio_list = rdir(path);

    int pos_in_last_folder = 0;
    foreach (Dbaudiofile& file, audio_list) {
      if (file.path == filename)
	break;
      else
	++pos_in_last_folder;
    }

    list<string> dirs = audio_folders;

    foreach (string& folder, audio_folders)
      if (filename.find(folder) != string::npos) {
	current_path = path.substr(0, folder.size());
	path = path.substr(folder.size());
      }

    int pos;

    while ((pos = path.find('/')) != string::npos && path.size() != 1) {

      vector<Dbaudiofile> curlist = rdir(current_path);

      current_path += path.substr(0, pos+1);

      int pos_in_folder = 0;
      foreach (Dbaudiofile& file, curlist) {
	if (file.path == current_path)
	  break;
	else
	  ++pos_in_folder;
      }

      folders.push(std::make_pair(dirs, pos_in_folder));

      path = path.substr(pos+1);

      dirs.clear();
      dirs.push_back(current_path);
    }

    folders.push(std::make_pair(dirs, pos_in_last_folder));

    loaded_correctly = true;

    return true;

  } else {
    delete q;
    db_mutex.leaveMutex();
    return false;
  }
}

void GraphicalAudio::reenter(int db_id)
{
  if (change_dir_to_id(db_id))
    mainloop(ADD);
}

vector<std::pair<string, int> > GraphicalAudio::gen_search_list(const string& search_word)
{
  string parent = "", audio_parent = "";

  if (input_master->current_saved_map() == "playlist") {

    string ids = "";
    foreach (Dbaudiofile& file, playlist)
      ids += conv::itos(file.db_id) + ", ";

    ids = ids.substr(0, ids.size()-2);

    parent = "id in (" + ids + ") AND";
    audio_parent = "filename in (SELECT filename FROM Folders WHERE id in (" + ids + ")) AND";

  } else if (input_master->current_saved_map() == "audio" && search_depth == dgettext("mms-audio", "current folder")) {
    string parent_sql_str = "SELECT DISTINCT id FROM %t WHERE";

    list<string> current_folders = folders.top().first;
    bool first = true;
    foreach (string& folder, current_folders) {
      if (first) {
	parent_sql_str += " filename='" + folder + "'";
	first = false;
      } else
	parent_sql_str += " OR filename='" + folder + "'";
    }

    string parent_ids = "";

    db_mutex.enterMutex();

    SQLQuery *q0 = db.query("Folders", parent_sql_str.c_str());

    if (q0) {
      for (int i = 0; i < q0->numberOfTuples(); ++i) {
	SQLRow &row = (*q0)[i];
	if (i > 0)
	  parent_ids += ", " + row["id"];
	else
	  parent_ids += row["id"];
      }
    }
    delete q0;

    db_mutex.leaveMutex();

    parent = "parent in (" + parent_ids + ") AND";
    audio_parent = "filename in (SELECT filename FROM Folders WHERE parent in (" + parent_ids + ")) AND";
  }

  vector<std::pair<string, int> > result;

  vector<int> ids_already_found;

  db_mutex.enterMutex();

  // folders
  SQLQuery *q = db.query("Folders", ("SELECT DISTINCT id, parent, filename, name, is_folder FROM %t WHERE " + parent + " fuzzycmp('" + search_word + "', lname, 1)").c_str());

  if (q) {

    for (int i = 0; i < q->numberOfTuples(); ++i) {

      SQLRow &row = (*q)[i];

      string out;

      if (row["is_folder"] == "1") {

	bool result_ok = false;

	if (search_depth == dgettext("mms-audio", "current folder")) {
	  list<string> current_folders = folders.top().first;
	  foreach (string& folder, current_folders) {
	    if (row["filename"].find(folder) != string::npos) {
	      result_ok = true;
	      break;
	    }
	  }
	} else
	  result_ok = true;

	if (result_ok && row["parent"] != "0") {
	  out = row["name"];
	  result.push_back(std::make_pair(string_format::convert(out), conv::atoi(row["id"])));
	}

      } else {

	SQLQuery *q3 = db.query("Audio", ("SELECT * FROM %t WHERE filename='" + string_format::escape_db_string(row["filename"]) + "'").c_str());
	if (q3 && q3->numberOfTuples() > 0) {
	  assert(q3->numberOfTuples() == 1);
	  SQLRow &c_row = (*q3)[0];
	  SQLQuery *q2 = db.query("Artist", ("SELECT name FROM %t WHERE id='" + c_row["Artist"] + "'").c_str());
	  if (q2 && q2->numberOfTuples() > 0) {
	    SQLRow &c_row2 = (*q2)[0];
	    out += c_row2["name"];
	  }
	  delete q2;

	  q2 = db.query("Album", ("SELECT name FROM %t WHERE id='" + c_row["Album"] + "'").c_str());
	  if (q2 && q2->numberOfTuples() > 0) {
	    SQLRow &c_row2 = (*q2)[0];
	    out += " - " + c_row2["name"];
	  }
	  delete q2;

	  out += " - " + c_row["Title"];

	  if (out == " -  - ")
	    out = row["name"];

	  result.push_back(std::make_pair(string_format::convert(out), conv::atoi(row["id"])));
	  ids_already_found.push_back(conv::atoi(row["id"]));
	}
	delete q3;
      }
    }
  }
  delete q;

  // title
  q = db.query("Audio", ("SELECT DISTINCT * FROM %t WHERE " + audio_parent + " fuzzycmp('" + search_word + "', lTitle, 1)").c_str());

  if (q) {
    for (int i = 0; i < q->numberOfTuples(); ++i) {

      SQLRow &row = (*q)[i];

      string out = "";

      SQLQuery *q2 = db.query("Artist", ("SELECT name FROM %t WHERE id='" + row["Artist"] + "'").c_str());
      if (q2 && q2->numberOfTuples() > 0) {
	SQLRow &c_row = (*q2)[0];
	out += c_row["name"];
      }
      delete q2;

      q2 = db.query("Album", ("SELECT name FROM %t WHERE id='" + row["Album"] + "'").c_str());
      if (q2 && q2->numberOfTuples() > 0) {
	SQLRow &c_row = (*q2)[0];
	out += " - " + c_row["name"];
      }
      delete q2;

      out += " - " + row["Title"];

      int id = 0;

      SQLQuery *q3 = db.query("Folders", ("SELECT id, name FROM %t WHERE filename='" + string_format::escape_db_string(row["filename"]) + "'").c_str());
      if (q3 && q3->numberOfTuples() > 0) {
	assert(q3->numberOfTuples() == 1);
	SQLRow &c_row = (*q3)[0];
	id = conv::atoi(c_row["id"]);
	if (out == " -  - ")
	  out = c_row["name"];
      } else
	assert(false);

      delete q3;

      // check if already found
      bool found_match = false;
      foreach (int id_already_found, ids_already_found)
	if (id == id_already_found) {
	  found_match = true;
	  break;
	}

      if (found_match)
	continue;

      result.push_back(std::make_pair(string_format::convert(out), id));
      ids_already_found.push_back(id);
    }
  }
  delete q;

  // artist
  q = db.query("Artist", ("SELECT id FROM %t WHERE fuzzycmp('" + search_word + "', lname, 1)").c_str());

  if (q && q->numberOfTuples() > 0) {

    string artist_ids;

    for (int i = 0; i < q->numberOfTuples(); ++i) {
      SQLRow &row = (*q)[i];

      artist_ids += row["id"] + ",";
    }

    artist_ids = artist_ids.substr(0, artist_ids.size()-1);

    SQLQuery *q2 = db.query("Audio", ("SELECT DISTINCT * FROM %t, Artist WHERE " + audio_parent + " Audio.Artist = Artist.id AND Artist in (" + artist_ids + ")").c_str());

    if (q2) {
      for (int j = 0; j < q2->numberOfTuples(); ++j) {
	SQLRow &row2 = (*q2)[j];

	int id;
	string name;

	SQLQuery *q4 = db.query("Folders", ("SELECT id FROM %t WHERE filename='" + string_format::escape_db_string(row2["filename"]) + "'").c_str());
	if (q4 && q4->numberOfTuples() > 0) {
	  assert(q4->numberOfTuples() == 1);
	  SQLRow &c_row = (*q4)[0];
	  id = conv::atoi(c_row["id"]);
	  name = c_row["name"];
	} else
	  assert(false);

	delete q4;

	// check if already found
	bool found_match = false;
	foreach (int id_already_found, ids_already_found)
	  if (id == id_already_found) {
	    found_match = true;
	    break;
	  }

	if (found_match)
	  continue;

	string out = row2["Artist.name"];

	SQLQuery *q3 = db.query("Album", ("SELECT name FROM %t WHERE id='" + row2["Album"] + "'").c_str());
	if (q3 && q3->numberOfTuples() > 0) {
	  SQLRow &c_row = (*q3)[0];
	  out += " - " + c_row["name"];
	}
	delete q3;

	out += " - " + row2["Title"];

	if (out == " -  - ")
	  out = name;

	result.push_back(std::make_pair(string_format::convert(out), id));
	ids_already_found.push_back(id);
      }
    }
    delete q2;
  }
  delete q;

  // album
  q = db.query("Album", ("SELECT id FROM %t WHERE fuzzycmp('" + search_word + "', lname, 1)").c_str());

  if (q && q->numberOfTuples() > 0) {

    string album_ids;

    for (int i = 0; i < q->numberOfTuples(); ++i) {
      SQLRow &row = (*q)[i];

      album_ids += row["id"] + ",";
    }

    album_ids = album_ids.substr(0, album_ids.size()-1);

    SQLQuery *q2 = db.query("Audio", ("SELECT DISTINCT * FROM %t, Album WHERE " + audio_parent + " Audio.Album = Album.id AND Album in (" + album_ids + ")").c_str());

    if (q2) {

      for (int j = 0; j < q2->numberOfTuples(); ++j) {
	SQLRow &row2 = (*q2)[j];

	int id;
	string name;

	SQLQuery *q4 = db.query("Folders", ("SELECT id FROM %t WHERE filename='" + string_format::escape_db_string(row2["filename"]) + "'").c_str());
	if (q4 && q4->numberOfTuples() > 0) {
	  assert(q4->numberOfTuples() == 1);
	  SQLRow &c_row = (*q4)[0];
	  id = conv::atoi(c_row["id"]);
	  name = c_row["name"];
	} else
	  assert(false);

	delete q4;

	// check if already found
	bool found_match = false;
	foreach (int id_already_found, ids_already_found)
	  if (id == id_already_found) {
	    found_match = true;
	    break;
	  }

	if (found_match)
	  continue;

	string out = "";

	SQLQuery *q3 = db.query("Artist", ("SELECT name FROM %t WHERE id='" + row2["Artist"] + "'").c_str());
	if (q3 && q3->numberOfTuples() > 0) {
	  SQLRow &c_row = (*q3)[0];
	  out += c_row["name"];
	}
	delete q3;

	out += " - " + row2["Album.name"] + " - " + row2["Title"];

	if (out == " -  - ")
	  out = name;

	result.push_back(std::make_pair(string_format::convert(out), id));
	ids_already_found.push_back(id);
      }
    }
    delete q2;
  }
  delete q;

  db_mutex.leaveMutex();

  return result;
}

int GraphicalAudio::check_search()
{
  return 1000; // ms
}

bool GraphicalAudio::search_is_graphical_view()
{
  return !(mode == PLAY || no_folders_in_dir(*files));
}

void GraphicalAudio::search_func()
{
  S_BusyIndicator::get_instance()->idle();

  ScreenUpdater *screen_updater = S_ScreenUpdater::get_instance();

  search_mode = true;
  search_help_offset = -1;

  print_marked_search_letter();

  list<string> empty;
  empty.push_back("");

  screen_updater->trigger.add(TriggerElement("search marker", empty, boost::bind(&GraphicalAudio::print_marked_search_letter, this),
					     boost::bind(&GraphicalAudio::clean_up_search_print, this)));
  screen_updater->timer.add(TimeElement("search marker", boost::bind(&GraphicalAudio::check_search_letter, this),
					boost::bind(&GraphicalAudio::print_marked_search_letter, this)));

  int pos = input_master->device->graphical_search<Dbaudiofile>(*files,
								boost::bind(&GraphicalAudio::print, this, _1),
								boost::bind(&GraphicalAudio::gen_search_list, this, _1),
								boost::bind(&GraphicalAudio::id_to_filename, this, _1),
								boost::bind(&GraphicalAudio::search_is_graphical_view, this),
								images_per_row, search_mode, search_str, lowercase_search_str,
								search_depth, offset, search_help_offset,
								S_BusyIndicator::get_instance(), mode == ADD);

  screen_updater->trigger.del("search marker");
  screen_updater->timer.del("search marker");

  if (pos != -1 && search_depth == dgettext("mms-audio", "current folder")) {
    if (mode == ADD)
      folders.top().second = pos;
    else
      playlist_pos_int = pos;
  } else if (pos != -1)
    change_dir_to_id(pos);
}

void GraphicalAudio::page_up()
{
  int jump = conf->p_jump();

  if (mode == ADD) {
    if (search_mode)
      jump = images_per_row*rows_search;
    else
      jump = images_per_row*rows;
  }

  if (files->size() > jump) {
    int new_pos = position_int()-jump;
    if (position_int() == 0)
      set_position_int(files->size()-1+new_pos);
    else if (new_pos < 0)
      set_position_int(0);
    else
      set_position_int(new_pos);
  }
}

void GraphicalAudio::page_down()
{
  int jump = conf->p_jump();

  if (mode == ADD) {
    if (search_mode)
      jump = images_per_row*rows_search;
    else
      jump = images_per_row*rows;
  }

  if (files->size() > jump) {
    if (position_int() > (files->size() - jump) && position_int() != (files->size()-1))
      set_position_int(files->size()-1);
    else
      set_position_int((position_int()+jump)%files->size());
  }
}

string GraphicalAudio::mainloop(AudioMode new_mode)
{
  BusyIndicator *busy_indicator = S_BusyIndicator::get_instance();
  Shutdown *sd = S_Shutdown::get_instance();

  search_top_size = 24;

  visible = true;

  mainloop_top(new_mode);

  bool exit_fall_through = false;

  Input input;

  bool update_needed = true;

  render->device->animation_section_begin();
  int curlayer = render->device->get_current_layer();
  render->device->switch_to_layer(curlayer+1);
  render->device->reset_layout_attribs_nowait();
  if (new_mode == !Audio::PLAY){
    render->device->animation_zoom(0,0,1,1,20,curlayer+1);
    render->device->animation_fade(0,1,40,curlayer+1);
    render->device->animation_fade(1,0,40,curlayer);
  } else{
    render->device->animation_zoom(3,3,1,1,40,curlayer+1);
    render->device->animation_fade(0,1,50,curlayer+1);
    render->device->animation_fade(1,0,50,curlayer);
  }

  render->device->animation_section_end();

  while (!exit_loop)
    {
      if (reload_dirs) {
	reload_current_dirs();
	update_needed = true;
	reload_dirs = false;
      }

      if (update_needed) {
	print_modifying(*files);
	print_lcd_menu();
       }

      input = input_master->device->get_input_busy_wrapped(busy_indicator);

      if (sd->is_enabled()) {
	sd->cancel();
	continue;
      }

      update_needed = true;

      if (input.key == "touch_input") {
	S_Touch::get_instance()->run_callback();
	continue;
      }

      // general
      if (input.command == "page_up")
 	{
	  page_up();
 	}
      else if (input.command == "page_down")
 	{
	  page_down();
 	}
      else if (input.command == "second_action")
	{
	  secondary_menu();
	}
      else if (input.command == "search")
	{
	  search_func();
	}
      else if (input.command == "options")
	{
	  options();
	}
      else if (input.command == "cover" && vector_lookup(*files, position_int()).type == "dir")
	{
	  choose_cover();
	}
      else if (input.command == "info" && input.mode == "graphical")
	{
	  print_information();
	}
      // playlist
      else if (mode == PLAY && input.command == "move_up")
	{
	  move_up();
	}
      else if (mode == PLAY && input.command == "move_down")
	{
	  move_down();
	}
      else if (mode == PLAY && input.command == "delete")
	{
	  delete_track();
	}
      else if (mode == PLAY && input.command == "queue")
	{
	  queue_track();
	}
      else if (mode == PLAY && input.command == "clear_list")
	{
	  clear_playlist();
	}
      else if (mode == PLAY && input.command == "save_playlist")
	{
	  save_playlist_func();
	}
      else if (mode == PLAY && input.command == "action")
	{
	  play_track();
	}
      else if (mode == PLAY && input.command == "startmenu" && input.mode == "general")
	{
	  exit_loop = true;
	  exit_fall_through = true;
	}
      else if (mode == ADD && input.command == "startmenu" && input.mode == "general")
	{
	  exit();
	}
      // audio-add and no folders
      else if (mode == ADD && !no_folders_in_dir(*files)) {
	if (input.command == "prev")
	  {
	    if (files->size() > images_per_row) {
	      for (int i = 0; i < images_per_row; ++i) {
		if (folders.top().second == 0) {
		  folders.top().second = files->size()-1;
		  i += images_per_row-files->size()%images_per_row;
		}
		else
		  folders.top().second -= 1;
	      }
	    }
	  }
	else if (input.command == "next")
	  {
	    if (files->size() > images_per_row) {
	      for (int i = 0; i < images_per_row; ++i) {
		if (folders.top().second == files->size()-1 && i == 0) {
		  folders.top().second = 0;
		  break;
		} else if (folders.top().second == files->size()-1) {
		  break;
		} else {
		  folders.top().second += 1;
		}
	      }
	    }
	  }
	else if (input.command == "left")
	  {
	    if (folders.top().second != 0)
	      folders.top().second -= 1;
	    else
	      folders.top().second = files->size() - 1;
	  }
	else if (input.command == "right")
	  {
	    folders.top().second = (folders.top().second+1)%files->size();
	  }
	else if (input.command == "back")
	  {
	    one_up();
	  }
	else if (input.command == "add_dir")
	  {
	    if (vector_lookup(*files, position_int()).type == "dir")
	      add();
	  }
	else if (input.command == "action")
	  {
	    if (vector_lookup(*files, position_int()).type == "dir")
	      enter_dir();
	    else
	      add();
	  }
	else if (input.command == "add_all")
	  {
	    add_all();
	  }
  	else if (input.command == "play_now")
	  {
	    if (audio_conf->p_audio_mode_playnow() &&
		vector_lookup(*files, position_int()).type == "dir")
	      play_now();
	  }
      } else {
	// general and no folders
	if (input.command == "prev")
	  {
	    if (position_int() != 0)
	      set_position_int(position_int()-1);
	    else
	      set_position_int(files->size()-1);
	  }
	else if (input.command == "next")
	  {
	    set_position_int((position_int()+1)%files->size());
	  }
	else if (input.command == "back" && input.mode == "general")
	  {
	    one_up();
	  }
	// audio-add and no folders
	else if (mode == ADD && input.command == "action")
	  {
	    add();
	  }
	else if (mode == ADD && input.command == "add_all")
	  {
	    add_all();
	  }
	else if (mode == ADD && input.command == "play_now")
	  {
	    if (vector_lookup(*files, position_int()).type != "dir")
		play_track_now();
	  }
      }
      update_needed = !global->check_commands(input);
    }

  render->device->animation_section_begin();
  render->device->layer_active_no_wait(false);
  render->device->switch_to_layer(curlayer);
  render->device->reset_layout_attribs_nowait();
  render->device->animation_fade(0,1,80,curlayer);
  render->device->animation_section_end();

  if (!exit_fall_through)
    exit_loop = false;

  if (in_playlist)
    in_playlist = false;

  visible = false;

  return "";
}

void GraphicalAudio::one_up()
{
  if (mode == ADD && folders.size() > 1)
    leave_dir();
  else
    exit();
}

string GraphicalAudio::get_cover(const Simplefile& file)
{
  string path = "";

  if (file.type == "media-track") {
    path = file.media_id;
  } else {
    if (file.type == "dir")
      path = string_format::unique_folder_name(file.path);
    else {
      string::size_type dir_seperator = file.path.rfind('/');
      if (dir_seperator != string::npos)
	path = string_format::unique_folder_name(file.path.substr(0, dir_seperator+1));
    }
  }

  string cover_path = "";

  db_mutex.enterMutex();

  SQLQuery *q = db.query("Covers", ("SELECT Cover FROM %t WHERE Path='" + string_format::escape_db_string(path) + "'").c_str());
  if (q && q->numberOfTuples() > 0) {
    SQLRow &row = (*q)[0];
    cover_path = row["Cover"];
  }
  delete q;

  db_mutex.leaveMutex();

  return cover_path;
}

void GraphicalAudio::secondary_menu_in_add(ExtraMenu &em)
{
  bool is_dir = (vector_lookup(*files, position_int()).type == "dir");

  if (is_dir) {
    em.add_item(ExtraMenuItem(dgettext("mms-audio", "Enter directory"), input_master->find_shortcut("action"),
			      boost::bind(&GraphicalAudio::enter_dir, this)));

    em.add_item(ExtraMenuItem(dgettext("mms-audio", "Add directory to playlist"), input_master->find_shortcut("add_dir"),
			      boost::bind(&GraphicalAudio::add, this)));

    em.add_item(ExtraMenuItem(dgettext("mms-audio", "Choose cover for folder"), input_master->find_shortcut("cover"),
			      boost::bind(&GraphicalAudio::choose_cover, this)));

  } else {
    em.add_item(ExtraMenuItem(dgettext("mms-audio", "Add track to playlist"), input_master->find_shortcut("action"),
			      boost::bind(&GraphicalAudio::add, this)));
  }

  if (folders.size() > 1)
    em.add_item(ExtraMenuItem(dgettext("mms-audio", "Go up one directory"), input_master->find_shortcut("back"),
			      boost::bind(&GraphicalAudio::leave_dir, this)));
  else
    em.add_item(ExtraMenuItem(dgettext("mms-audio", "Return to startmenu"), input_master->find_shortcut("back"),
			      boost::bind(&GraphicalAudio::exit, this)));

  em.add_item(ExtraMenuItem(dgettext("mms-audio", "Show Lyrics"), input_master->find_shortcut("show_lyrics"),
			    boost::bind(&GraphicalAudio::show_lyrics, this)));
}

void GraphicalAudio::secondary_menu_commands(ExtraMenu &em)
{
  if (vector_lookup(*files, position_int()).type != "dir")
    em.add_item(ExtraMenuItem(dgettext("mms-audio", "Print information"), input_master->find_shortcut("info"),
			      boost::bind(&GraphicalAudio::print_information, this)));

  if (opts.lyrics() == dgettext("mms-audio", "yes") && audio_state->p_playing())
    em.add_item(ExtraMenuItem(dgettext("mms-audio", "Show Lyrics"), input_master->find_shortcut("show_lyrics"),
			      boost::bind(&GraphicalAudio::show_lyrics, this)));
}

void GraphicalAudio::enter_dir()
{
  if (vector_lookup(*files, position_int()).type == "dir") {
    string temp_cur_dir = vector_lookup(*files, position_int()).path;

    int nr_files = 0;
    for (file_iterator<file_t, default_order> i (temp_cur_dir); i != i.end (); i.advance(false))
      if (!isDirectory(i->getName()) && check_type(i->getName(), audio_conf->p_filetypes_a()) != emptyMyPair)
	++nr_files;

    ScreenUpdater *screen_updater = S_ScreenUpdater::get_instance();

    screen_updater->status_progressbar = 0;
    screen_updater->total_progressbar = nr_files;
    screen_updater->header = dgettext("mms-audio", "Extracting metadata");

    screen_updater->timer.add(TimeElement("progressbar", &check_progressbar, &print_progressbar));

    vector<Dbaudiofile> templist = rdir(temp_cur_dir);

    screen_updater->timer.del("progressbar");
    cleanup_progressbar();

    if (templist.size() != 0) {

      list<string> templs;
      if (temp_cur_dir[temp_cur_dir.size()-1] != '/')
	templs.push_back(temp_cur_dir + '/');
      else
	templs.push_back(temp_cur_dir);
      folders.push(std::make_pair(templs, 0));
      audio_list = templist;
      files = &audio_list;

      S_BackgroundUpdater::get_instance()->run_once(boost::bind(&GraphicalAudio::check_for_changes, this));

    } else
      DialogWaitPrint pdialog(dgettext("mms-audio", "Folder is empty"), 1000);
  }
}

void GraphicalAudio::check_for_changes()
{
  bool reload = false;

  list<string> top_folders = folders.top().first;

  foreach (string& dir, top_folders)
    if (reload_dir(dir))
      reload = true; // no break since we want to update the db for all folders

  reload_dirs = reload;
}

void GraphicalAudio::print_audiotrack_helper(int &x_pos)
{
  string pic = get_cover(audio_state->p->p_cur_nr());

  if (!pic.empty()) {
    x_pos = 70;

    render->create_scaled_image_wrapper(pic, 57, 63);
    int pic_y_pos = conf->p_v_res()-65;
    if (conf->p_h_res() == 720)
      pic_y_pos += 2;
    audio_state->playback.add(new PFObj(pic, 7 + global->get_playback_offset(), pic_y_pos, 57, 63, false, 0));
  }

  x_pos += global->get_playback_offset();
}

void GraphicalAudio::get_audiotrack_info(string& buffer, string& artist, string& album, string& title)
{
  // the extracted information is more reliant for VBR mp3 than the audio
  // player, same for artist, album and title
  if (!audio_state->p->p_cur_nr().path.empty() && audio_state->p->p_cur_nr().type != "web" &&
      audio_state->p->p_cur_nr().type != "media-file" && audio_state->p->p_cur_nr().type != "media-track") {

    Dbaudiofile dbaf(audio_state->p->p_cur_nr().path);
    buffer = format_time(audio_state->p->p_cur_time(), dbaf.length != 0 ? dbaf.length : audio_state->p->p_total_time());
    artist = string_format::convert(dbaf.artist, true);
    album = string_format::convert(dbaf.album, true);
    title = string_format::convert(dbaf.title, true);

  } else {
    if (audio_state->p->p_cur_nr().type == "web")
      buffer = format_time(audio_state->p->p_cur_time());
    else
      buffer = format_time(audio_state->p->p_cur_time(), audio_state->p->p_total_time());
    artist = audio_state->p->p_artist();
    album = audio_state->p->p_album();
    title = audio_state->p->p_title();
  }
}

void GraphicalAudio::print_audio_screensaver_helper(int &x_pos, int &y, int max_x_pic,
						    int max_y_pic, bool vert_center)
{
  string pic = get_cover(audio_state->p->p_cur_nr());

  if (!pic.empty()) {
    render->create_scaled_image_wrapper(pic, max_x_pic, max_y_pic);
    PFObj *p = new PFObj(pic, 60, 60, max_x_pic, max_y_pic, false, 2);
    if (vert_center)
      y = (conf->p_v_res()-p->real_h)/2;
    p->real_y = p->y = y;
    audio_state->playback_fullscreen.add(new RObj(55, y-5, p->w+10, p->h+10, 255, 255, 255, 255, 1));
    audio_state->playback_fullscreen.add(p);
  } else
    x_pos = 100;
}

void GraphicalAudio::print_choose_cover(const vector<string>& covers, int cover_pos, Overlay& o)
{
  render->wait_and_aquire();

  S_Touch::get_instance()->clear();

  if (o.elements.size() > 0)
    o.partial_cleanup();

  int x, y, max_x = conf->p_h_res()-250, max_y = 90;

  int cover_width = image_width+25;
  int cover_height = image_height+15;

  int covers_per_row = (max_x-30)/cover_width;
  int image_rows = (conf->p_v_res()-(2*max_y))/cover_height;

  int max = std::min(covers_per_row*image_rows, int(covers.size()));

  int box_height = cover_height*(((max % covers_per_row == 0) ? 0 : 1) + max/covers_per_row)+30+50;

  x = (conf->p_h_res()-(max_x+30))/2;
  y = (conf->p_v_res()-(box_height))/2;

  o.add(new RObj(x-15+2, y-15+2, max_x+30+1, box_height+1, 0, 0, 0, 200, 0));
  o.add(new RObj(x-15, y-15, max_x+30, box_height,
		 themes->dialog_background1, themes->dialog_background2,
		 themes->dialog_background3, themes->dialog_background4, 1));

  string header = dgettext("mms-audio", "choose a cover");

  int x_size = string_format::calculate_string_width(header, "Vera/20");

  o.add(new TObj(header, "Vera/20", (conf->p_h_res()-x_size)/2, y-5,
		 themes->dialog_font1, themes->dialog_font2, themes->dialog_font3, 3));
  y += 50;

  // print the range
  int pos_in_list = cover_pos;
  int screen_pos, start;

  if (covers.size() > image_rows*covers_per_row) {
    if (pos_in_list < std::ceil(double(image_rows)/2.0)*covers_per_row) {
      screen_pos = pos_in_list;
      start = 0;
    } else {
      screen_pos = pos_in_list%covers_per_row+covers_per_row;
      if (screen_pos == covers_per_row-1)
	screen_pos = 2*covers_per_row-1;
      start = pos_in_list-screen_pos;
    }
  } else {
    screen_pos = pos_in_list;
    start = 0;
  }

  x = (conf->p_h_res()-covers_per_row*cover_width)/2;
  int x_start = x;

  for (int j = 0; j < image_rows*covers_per_row && j < covers.size(); ++j) {

    string pic = vector_lookup(covers, start);

    bool border = false;

    if (screen_pos == j)
      border = true;

    if (!file_exists(render->image_get_fullpath(pic, image_width, image_height))) {
      render->create_scaled_image_wrapper(pic, image_width, image_height);
      // to insert the image into the transmap so that it won't fail when
      // we draw the screen again
      bool tmp;
      render->image_dimensions(render->image_get_fullpath(pic, image_width, image_height), &tmp);
    }

    PFObj *p = new PFObj(pic, x, y, image_width, image_height, border, 3);

    p->x = p->real_x = x + (cover_width - (p->w + 10))/2;
    p->y = p->real_y = y + (cover_height - (p->h + 10))/2;

    if (p->border) {
      p->real_x -= 5;
      p->real_y -= 5;
    } else {
      p->real_x -= 1;
      p->real_y -= 1;
    }

    o.add(p);

    S_Touch::get_instance()->register_area(TouchArea(rect(p->x, p->y, p->w, p->h), p->layer, boost::bind(&GraphicalAudio::find_cover_element_and_do_action, this, pic)));

    x += cover_width;

    if (start == covers.size()-1) {
      if ((j+1)%covers_per_row != 0)
	j += (covers_per_row-(j+1)%covers_per_row);

      x = x_start;
      y += cover_height;

      start = 0;
      continue;

    } else
      ++start;

    if ((j+1)%covers_per_row == 0) {
      y += cover_height;
      x = x_start;
    }
  }

  render->draw_and_release("choose cover screen", true);
}

void GraphicalAudio::choose_cover()
{
  // find all covers available
  covers.clear();
  cover_pos = 0;

  MyPair filetype;

  libfs_set_folders_first(opts.dir_order());

  for (file_iterator<file_t, default_order> i (vector_lookup(*files, position_int()).path);
       i != i.end (); ++i)
    {
      filetype = check_type(i->getName(), conf->p_filetypes_p());

      if (filetype != emptyMyPair)
	covers.push_back(i->getName());
    }

  if (covers.size() == 0) {
    DialogWaitPrint pdialog(dgettext("mms-audio", "No covers found"), 2000);
    return;
  }

  // display a list and let the user select one
  S_BusyIndicator::get_instance()->idle();

  ScreenUpdater *screen_updater = S_ScreenUpdater::get_instance();
  screen_updater->timer.deactivate("audio_fullscreen");

  Overlay overlay = Overlay("dialog");

  Input input;

  int max_x = conf->p_h_res()-250;
  int cover_width = image_width+25;
  int covers_per_row = (max_x-30)/cover_width;

  while (!exit_choose_cover_loop) {

    print_choose_cover(covers, cover_pos, overlay);

    input = input_master->device->get_input();

    if (input.key == "touch_input") {
      S_Touch::get_instance()->run_callback();
      continue;
    }

    if (input.command == "back" && input.mode == "general")
      {
	exit_choose_cover_loop = true;
      }
    else if (input.command == "left")
      {
	if (cover_pos != 0)
	  --cover_pos;
	else
	  cover_pos = covers.size() - 1;
      }
    else if (input.command == "right")
      {
	if (cover_pos != covers.size()-1)
	  ++cover_pos;
	else
	  cover_pos = 0;
      }
    else if (input.command == "prev")
      {
	if (covers.size() > covers_per_row) {
	  cover_pos = cover_pos - covers_per_row;
	  if (cover_pos < 0)
	    cover_pos = covers.size() - 1 + cover_pos;
	}
      }
    else if (input.command == "next")
      {
	cover_pos = (cover_pos + covers_per_row) % covers.size();
      }
    else if (input.command == "action")
      {
	use_cover();
      }
  }

  render->wait_and_aquire();
  overlay.cleanup();
  render->image_mut.leaveMutex();

  exit_choose_cover_loop = false;

  screen_updater->timer.activate("audio_fullscreen");
}

void GraphicalAudio::find_cover_element_and_do_action(const string& pic)
{
  int p = 0;
  foreach (string& cover, covers) {
    if (cover == pic) {
      cover_pos = p;
      break;
    } else
      ++p;
  }

  use_cover();
}

void GraphicalAudio::use_cover()
{
  string cur_dir = string_format::unique_folder_name(vector_lookup(*files, position_int()).path);

  db_mutex.enterMutex();
  db.execute(("UPDATE Covers SET Cover = '" + string_format::escape_db_string(vector_lookup(covers, cover_pos)) + "' WHERE Path = '" + string_format::escape_db_string(cur_dir) + "'").c_str());
  db_mutex.leaveMutex();

#if 0
  if (!file_exists(render->image_get_fullpath(cover, image_width, image_height))) {
    render->create_scaled_image_wrapper(cover, image_width, image_height);
    // to insert the image into the transmap so that it won't fail when
    // we draw the screen again
    bool tmp;
    render->image_dimensions(render->image_get_fullpath(cover, image_width, image_height), &tmp);
  }
#endif

  exit_choose_cover_loop = true;
}

void GraphicalAudio::print_information()
{
  if (isDirectory(vector_lookup(*files, position_int()).path))
    return;

  visible = false;

  S_BusyIndicator::get_instance()->idle();

  printing_information = true;

  render->prepare_new_image();

  render->current.add(new PObj(themes->audio_background, 0, 0, 0, SCALE_FULL));

  render->current.add(new PObj(themes->startmenu_music_dir, 25, 10, 2, NOSCALING));

  if (themes->show_header) {
    render->current.add(new TObj(dgettext("mms-audio", "Audio - Information"), "Vera/28", 100, 10,
				 themes->audio_header_font1, themes->audio_header_font2,
				 themes->audio_header_font3, 2));
  }

  int y = 100;

  Dbaudiofile dbaf = vector_lookup(*files, position_int());
  // lookup information each time see this information as the rating of the
  // song might have changed
  if (dbaf.type != "media-track" && dbaf.type != "media-file" && dbaf.type != "web")
    dbaf = Dbaudiofile(dbaf.path);

  string pic = get_cover(vector_lookup(*files, position_int()));

  int max_x_pic = 240, max_y_pic = 266;

  if (conf->p_h_res() == 1024) {
    max_x_pic = 300;
    max_y_pic = 333;
  }

  int x = max_x_pic + 35 + 35;
  if (pic.empty())
    x = 50;

  if (!pic.empty()) {
    render->create_scaled_image_wrapper(pic, max_x_pic, max_y_pic);
    render->current.add(new PFObj(pic, 35, y+5, max_x_pic, max_y_pic, true, 3));
  }

  int w, first_column_width = 0;
  w = string_format::calculate_string_width(dgettext("mms-audio", "Artist:"), "Vera/18");
  first_column_width = std::max(w, first_column_width);
  w = string_format::calculate_string_width(dgettext("mms-audio", "Album:"), "Vera/18");
  first_column_width = std::max(w, first_column_width);
  w = string_format::calculate_string_width(dgettext("mms-audio", "Title:"), "Vera/18");
  first_column_width = std::max(w, first_column_width);
  w = string_format::calculate_string_width(dgettext("mms-audio", "Year:"), "Vera/18");
  first_column_width = std::max(w, first_column_width);
  w = string_format::calculate_string_width(dgettext("mms-audio", "Bitrate:"), "Vera/18");
  first_column_width = std::max(w, first_column_width);
  w = string_format::calculate_string_width(dgettext("mms-audio", "Length:"), "Vera/18");
  first_column_width = std::max(w, first_column_width);
#ifdef use_imms
  w = string_format::calculate_string_width(dgettext("mms-audio", "Rating:"), "Vera/18");
  first_column_width = std::max(w, first_column_width);
#endif
  w = string_format::calculate_string_width(dgettext("mms-audio", "Track:"), "Vera/18");
  first_column_width = std::max(w, first_column_width);

  first_column_width += 10;

  string artist = string_format::convert(dbaf.artist, true);
  string_format::format_to_size(artist, "Vera/18", conf->p_h_res()-(x+first_column_width)-25, true);
  render->current.add(new TObj(dgettext("mms-audio", "Artist:"), "Vera/18", x, y, themes->audio_font1,
			       themes->audio_font2, themes->audio_font3, 3));
  render->current.add(new TObj(artist, "Vera/18",
			       x+first_column_width, y,
			       themes->audio_font1, themes->audio_font2,
			       themes->audio_font3, 3));
  y+=28;

  string album = string_format::convert(dbaf.album, true);
  string_format::format_to_size(album, "Vera/18", conf->p_h_res()-(x+first_column_width)-25, true);
  render->current.add(new TObj(dgettext("mms-audio", "Album:"), "Vera/18", x, y,
			       themes->audio_font1, themes->audio_font2,
			       themes->audio_font3, 3));
  render->current.add(new TObj(album, "Vera/18",
			       x+first_column_width, y,
			       themes->audio_font1, themes->audio_font2,
			       themes->audio_font3, 3));
  y+=28;

  string title = string_format::convert(dbaf.title, true);
  string_format::format_to_size(title, "Vera/18", conf->p_h_res()-(x+first_column_width)-25, true);
  render->current.add(new TObj(dgettext("mms-audio", "Title:"), "Vera/18", x, y,
			       themes->audio_font1, themes->audio_font2,
			       themes->audio_font3, 3));
  render->current.add(new TObj(title, "Vera/18",
			       x+first_column_width, y,
			       themes->audio_font1, themes->audio_font2,
			       themes->audio_font3, 3));
  y+=28;

  render->current.add(new TObj(dgettext("mms-audio", "Year:"), "Vera/18", x, y,
			       themes->audio_font1, themes->audio_font2,
			       themes->audio_font3, 3));
  render->current.add(new TObj(conv::itos(dbaf.year), "Vera/18",
			       x+first_column_width, y,
			       themes->audio_font1, themes->audio_font2,
			       themes->audio_font3, 3));
  y+=28;

  render->current.add(new TObj(dgettext("mms-audio", "Bitrate:"), "Vera/18", x, y,
			       themes->audio_font1, themes->audio_font2,
			       themes->audio_font3, 3));
  render->current.add(new TObj(conv::itos(dbaf.bitrate), "Vera/18",
			       x+first_column_width, y,
			       themes->audio_font1, themes->audio_font2,
			       themes->audio_font3, 3));
  y+=28;

  std::ostringstream time;

  if (remaining_seconds(dbaf.length) < 10)
    time << seconds_to_minutes(dbaf.length) << ":0" << remaining_seconds(dbaf.length);
  else
    time << seconds_to_minutes(dbaf.length) << ":" << remaining_seconds(dbaf.length);

  render->current.add(new TObj(dgettext("mms-audio", "Length:"), "Vera/18", x, y,
			       themes->audio_font1, themes->audio_font2,
			       themes->audio_font3, 3));
  render->current.add(new TObj(time.str(), "Vera/18",
			       x+first_column_width, y,
			       themes->audio_font1, themes->audio_font2,
			       themes->audio_font3, 3));
  y+=28;

#ifdef use_imms
  std::ostringstream rating;
  if (dbaf.rating != -1)
    rating << dbaf.rating;
  else
    rating << dgettext("mms-audio", "not rated yet");

  render->current.add(new TObj(dgettext("mms-audio", "Rating:"), "Vera/18", x, y,
			       themes->audio_font1, themes->audio_font2,
			       themes->audio_font3, 3));
  render->current.add(new TObj(rating.str(), "Vera/18", x+first_column_width, y,
			       themes->audio_font1, themes->audio_font2,
			       themes->audio_font3, 3));
  y+=28;
#endif

  render->current.add(new TObj(dgettext("mms-audio", "Track:"), "Vera/18", x, y,
			       themes->audio_font1, themes->audio_font2,
			       themes->audio_font3, 3));
  render->current.add(new TObj(conv::itos(dbaf.track), "Vera/18",
			       x+first_column_width, y,
			       themes->audio_font1, themes->audio_font2,
			       themes->audio_font3, 3));
  y+=28;

  render->draw_and_release("Audio information");

  input_master->device->get_input();

  printing_information = false;

  visible = true;
}

bool GraphicalAudio::file_exists_in_db(const string& file)
{
  return (db_id(file, false) != -1);
}

int GraphicalAudio::db_id(const string& file, bool is_dir)
{
  int i = -1;

  string good_file;

  if (is_dir)
    good_file = string_format::unique_folder_name(file);
  else
    good_file = file;

  db_mutex.enterMutex();
  SQLQuery *q = db.query("Folders", ("SELECT id FROM %t WHERE filename='" + string_format::escape_db_string(good_file) + "'").c_str());
  if (q && q->numberOfTuples() > 0) {
    SQLRow &row = (*q)[0];
    i = conv::atoi(row["id"]);
  }
  delete q;
  db_mutex.leaveMutex();

  return i;
}

string GraphicalAudio::get_name_from_file(const Dbaudiofile& d)
{
  return d.to_string();
}

string GraphicalAudio::get_short_name_from_file(const Dbaudiofile& d)
{
  return d.short_to_string();
}

void GraphicalAudio::print_element(Dbaudiofile& dbaf, const Dbaudiofile& position,
				   int y, bool picture_dir)
{
  if (!dbaf.is_fetched() && dbaf.type != "media-track" && dbaf.type != "media-file" &&
      dbaf.type != "web") {
    Dbaudiofile tmp_dbaf(dbaf.path);
    tmp_dbaf.set_values(dbaf);
    dbaf = tmp_dbaf;
  }

  string name;

  if (mode == ADD && picture_dir && (search_depth == dgettext("mms-audio", "current folder") || !search_mode))
    name = dbaf.get_name_from_file_in_picture_dir();
  else
    name = get_name_from_file(dbaf);

  if (picture_dir && mode == ADD && (search_depth == dgettext("mms-audio", "current folder") || !search_mode)) {

    int x_end = 358;

    if (conf->p_h_res() == 1024)
      x_end = 418;

    if (conf->p_v_res() == 405)
      x_end = 250;

    string_format::format_to_size(name, "Vera/16", conf->p_h_res()-x_end-85, true);

    // touch
    PObj *p = new PObj(themes->audio_marked_small_left, 75, y, 2, NOSCALING);

    if (mode == ADD)
      S_Touch::get_instance()->register_area(TouchArea(rect(p->x, p->y, conf->p_h_res()-x_end-75, p->h), p->layer, boost::bind(&GraphicalAudio::find_element_and_do_action_add, this, dbaf)));
    else
      S_Touch::get_instance()->register_area(TouchArea(rect(p->x, p->y, conf->p_h_res()-x_end-75, p->h), p->layer, boost::bind(&GraphicalAudio::find_element_and_do_action_playlist, this, dbaf)));

    // shorter
    if (dbaf == position) {
      render->current.add(p);
      render->current.add(new PObj(themes->audio_marked_small, 75+25, y, conf->p_h_res()-x_end-75 - 50, 2));
      render->current.add(new PObj(themes->audio_marked_small_right, 75+25+conf->p_h_res()-x_end-75 - 50, y, 2, NOSCALING));
    }

    if (isDirectory(dbaf.path))
      name.append("/");

    render->current.add(new TObj(name, "Vera/16", 75+7, y-2, themes->audio_font1, themes->audio_font2,
				 themes->audio_font3, 3));

  } else {

    int width = conf->p_h_res()-250;
    if (mode == ADD)
      width += 15;

    string_format::format_to_size(name, "Vera/16", width, true);

    PObj *p = new PObj(themes->audio_marked_left, 70, y, 2, NOSCALING);

    if (mode == ADD)
      S_Touch::get_instance()->register_area(TouchArea(rect(p->x, p->y, conf->p_h_res()-75-70, p->h), p->layer, boost::bind(&GraphicalAudio::find_element_and_do_action_add, this, dbaf)));
    else
      S_Touch::get_instance()->register_area(TouchArea(rect(p->x, p->y, conf->p_h_res()-75-70, p->h), p->layer, boost::bind(&GraphicalAudio::find_element_and_do_action_playlist, this, dbaf)));

    if (dbaf == position) {
      render->current.add(p);
      render->current.add(new PObj(themes->audio_marked, 70+25, y, conf->p_h_res()-75-70-50, 2));
      render->current.add(new PObj(themes->audio_marked_right, 70+25+conf->p_h_res()-75-70-50, y, 2, NOSCALING));
    }

    bool printed = false;
    if (mode == PLAY)
      if (dbaf.id == audio_state->p->p_cur_nr().id && audio_state->p_playing()) {
	render->current.add(new TObj(name, "Vera/16", 75, y-2,
				     themes->audio_marked_font1, themes->audio_marked_font2,
				     themes->audio_marked_font3, 3));
	printed = true;
      }

    if (!printed) {

      if (isDirectory(dbaf.path))
	name.append("/");

      render->current.add(new TObj(name, "Vera/16", 75, y-2,
				   themes->audio_font1, themes->audio_font2,
				   themes->audio_font3, 3));
    }

    if (dbaf.length != 0) {
      char temp[512];
      snprintf(temp, 512, "%d:%02d", dbaf.length/60, dbaf.length%60);
      int x_size = string_format::calculate_string_width(temp, "Vera/16");
      render->current.add(new TObj(temp, "Vera/16", conf->p_h_res()-75-x_size, y-2,
				   themes->audio_font1, themes->audio_font2,
				   themes->audio_font3, 3));
    }

    // FIXME: inefficient
    int qpos = audio_state->queue_pos(dbaf);

    if (qpos != 0)
      render->current.add(new TObj(conv::itos(qpos), "Vera/16", conf->p_h_res()-165, y-2,
				   themes->audio_font1, themes->audio_font2,
				   themes->audio_font3, 3));
  }
}

void GraphicalAudio::print_modifying(vector<Dbaudiofile>& cur_files)
{
  if (mode == PLAY || no_folders_in_dir(cur_files))
    print_tracklist(cur_files);
  else
    print_graphical(cur_files);
}

bool GraphicalAudio::no_folders_in_dir(const vector<Dbaudiofile>& cur_files)
{
  foreach (const Dbaudiofile& file, cur_files)
    if (file.type == "dir")
      return false;

  return true;
}

void GraphicalAudio::options_reloaded()
{
  extracted_all_files = false;
}

string GraphicalAudio::get_cover_from_dir(const string& dir)
{
  string pic = "";

  vector<string> pictures_found;

  MyPair filetype;

  // FIXME: is this right? Do we still use .front-cover.jpg? Or just legacy?

  // check for special cover first
  if (file_exists(dir + "/.front-cover.jpg")) {
    pic = dir + "/.front-cover.jpg";
  } else {
    libfs_set_folders_first(opts.dir_order());

    // determine if there is a picture in this dir...
    for (file_iterator<file_t, default_order> i (dir); i != i.end (); i.advance(false))
      {
       filetype = check_type(i->getName(), conf->p_filetypes_p());

       if (filetype != emptyMyPair)
         pictures_found.push_back(i->getName());
      }

    if (!pictures_found.empty()) {
      pic = pictures_found.front();
      if (pictures_found.size() > 1) {
	foreach (string& picture, pictures_found)
	  if (string_format::lowercase(picture).rfind("front") != string::npos) {
	    pic = picture;
	    break;
	  }
      }
    }
  }

  return pic;
}

string GraphicalAudio::get_first_cover_in_dir(const string& dir)
{
  string cover = "";
  bool cover_found = false;

  string cur_dir = dir;

  list<string> dirs_to_try;

  while (!cover_found) {
    cover = get_cover_from_dir(cur_dir);

    if (!cover.empty()) {
      cover_found = true;
    }
    else {
      bool new_dir_found = false;

      libfs_set_folders_first(opts.dir_order());

      for (file_iterator<file_t, default_order> i (cur_dir); i != i.end (); i.advance(false)) {
       if (isDirectory(i->getName())) {
         if (!new_dir_found) {
           new_dir_found = true;
           cur_dir = i->getName();
         } else {
           dirs_to_try.push_back(i->getName());
         }
       }
      }
      if (!new_dir_found) {
       if (!dirs_to_try.empty()) {
         cur_dir = dirs_to_try.front();
         dirs_to_try.pop_front();
       } else
         break;
      }
    }
  }

  return cover;
}

void GraphicalAudio::print_dir(int cur_pos, int j, int x, int y, const Dbaudiofile& s)
{
  PObj *p;

  if (cur_pos == j)
    p = new PObj(themes->audio_dir_marked, x, y, 3, NOSCALING);
  else
    p = new PObj(themes->audio_dir, x, y, 3, NOSCALING);

  p->x = p->real_x = x + (image_width_all - p->w)/2;
  p->y = p->real_y = y + (image_height_all_eks_text - p->h)/2;

  render->current.add(p);

  S_Touch::get_instance()->register_area(TouchArea(rect(p->x, p->y, p->w, p->h), p->layer, boost::bind(&GraphicalAudio::find_element_and_do_action_add, this, s)));
}

void GraphicalAudio::find_element(const Dbaudiofile& audio_file)
{
  int p = 0;
  foreach (Dbaudiofile& f, *files) {
    if (f == audio_file) {
      set_position_int(p);
      break;
    } else
      ++p;
  }
}

void GraphicalAudio::find_element_and_do_action_playlist(const Dbaudiofile& audio_file)
{
  find_element(audio_file);

  if (search_mode) {
    input_master->device->got_touch_action();
    return;
  } else
    play_track();
}

void GraphicalAudio::find_element_and_do_action_add(const Dbaudiofile& audio_file)
{
  find_element(audio_file);

  if (search_mode) {
    input_master->device->got_touch_action();
    return;
  } else {
    if (vector_lookup(*files, position_int()).type == "dir")
      enter_dir();
    else
      add();
  }
}

void GraphicalAudio::print_graphical(vector<Dbaudiofile>& cur_files)
{
  render->prepare_new_image();

  render->current.add(new PObj(themes->audio_background, 0, 0, 0, SCALE_FULL));

  PObj *back = new PObj(themes->startmenu_music_dir, 25, 10, 2, NOSCALING);

  render->current.add(back);

  S_Touch::get_instance()->register_area(TouchArea(rect(back->x, back->y, back->w, back->h), back->layer, boost::bind(&GraphicalAudio::one_up, this)));

  if (themes->show_header) {

    string header = dgettext("mms-audio", "Audio Library");

    if (folders.size() > 1) {
      string top_folder = folders.top().first.front();

      if (top_folder[top_folder.size()-1] == '/')
	top_folder = top_folder.substr(0, top_folder.size()-1);

      assert(top_folder.rfind('/') != string::npos);
      header += " - " + top_folder.substr(top_folder.rfind('/')+1);
      string_format::format_to_size(header, "Vera/28", conf->p_h_res()-220, false);
    }

    TObj *t = new TObj(header, "Vera/28", 100, 5, themes->audio_header_font1, themes->audio_header_font2,
		       themes->audio_header_font3, 2);

    render->current.add(t);

    S_Touch::get_instance()->register_area(TouchArea(rect(t->x, t->y, t->w, t->h), t->layer, boost::bind(&GraphicalAudio::toggle_playlist, this)));
  }

  if (search_mode) {

    int y = 60;
    if (conf->p_h_res() == 720) {
      y -= 3;
    }

    if (offset == -1)
      render->current.add(new RObj(0, y, conf->p_h_res(), 45, 0, 0, 0, 215, 2));
    else
      render->current.add(new RObj(0, y, conf->p_h_res(), 45,
				   themes->general_search_rectangle_color1,
				   themes->general_search_rectangle_color2,
				   themes->general_search_rectangle_color3,
				   themes->general_search_rectangle_color4, 2));

    render->current.add(new PObj(themes->general_search, 47, y, 3, NOSCALING));

    int max_x = conf->p_h_res()-250;

    string sw = search_str;

    string s = dgettext("mms-audio", "Search: ");
    int x_size = string_format::calculate_string_width(s, "Vera/22");
    string_format::format_to_size(sw, "Vera/22", max_x-x_size, true);

    render->current.add(new TObj(s + sw, "Vera/22", 100, y+5,
				 themes->search_font1, themes->search_font2,
				 themes->search_font3, 3));

    render->current.add(new TObj(search_depth, "Vera/17", 50+max_x+25, y+10, themes->search_font1,
				 themes->search_font2, themes->search_font3, 3));
  }

  // print the range
  int screen_pos, start, pos_in_list, cur_rows, cur_image_height_all;

  if (search_mode) {
    if (offset == -1)
      pos_in_list = 0;
    else
      pos_in_list = offset;
    cur_rows = rows_search;
    cur_image_height_all = image_height_all_search;
  } else {
    pos_in_list = folders.top().second;
    cur_rows = rows;
    cur_image_height_all = image_height_all;
  }

  int nr_in_last_row = cur_files.size()%images_per_row;
  if (nr_in_last_row == 0)
    nr_in_last_row = images_per_row;

  if (cur_files.size() > cur_rows*images_per_row) {
    if (pos_in_list < std::ceil(double(cur_rows)/2.0)*images_per_row) {
      screen_pos = pos_in_list;
      start = 0;
    } else if (pos_in_list >= cur_files.size() - (images_per_row+nr_in_last_row)) { // last two rows
      start = cur_files.size()-((cur_rows-1)*images_per_row+nr_in_last_row);
      screen_pos = pos_in_list-start;
    } else {
      screen_pos = pos_in_list%images_per_row+images_per_row;
      if (screen_pos == images_per_row-1)
	screen_pos = 2*images_per_row-1;
      start = pos_in_list-screen_pos;
    }
  } else {
    screen_pos = pos_in_list;
    start = 0;
  }

  int x = (conf->p_h_res()-images_per_row*image_width_all)/2;
  int x_start = x;
  int y = y_start;

  if (search_mode) {
    if (conf->p_input() != "keyboard") {
      y += 65;
      if (conf->p_h_res() == 800)
	y += 25;
    } else
      y += 45;
  }

  for (int j = 0; j < cur_rows*images_per_row && j < cur_files.size(); ++j) {

    Dbaudiofile s = vector_lookup(cur_files, start);

    if (s.type == "dir") {

      string pic = get_cover(s);

      if (file_exists(pic)) {

	bool border = false;

	if (screen_pos == j)
	  border = true;

	if (!file_exists(render->image_get_fullpath(pic, image_width, image_height))) {
	  render->create_scaled_image_wrapper(pic, image_width, image_height);
	  // to insert the image into the transmap so that it won't fail when
	  // we draw the screen again
	  bool tmp;
	  render->image_dimensions(render->image_get_fullpath(pic, image_width, image_height), &tmp);
	}

	PFObj *p = new PFObj(pic, x, y, image_width, image_height, border, 3);

	p->x = p->real_x = x + (image_width_all - (p->w + 10))/2;
	p->y = p->real_y = y + (image_height_all_eks_text - (p->h + 10))/2;

	if (p->border) {
	  p->real_x -= 5;
	  p->real_y -= 5;
	} else {
	  p->real_x -= 1;
	  p->real_y -= 1;
	}

	S_Touch::get_instance()->register_area(TouchArea(rect(p->x, p->y, p->w, p->h), p->layer, boost::bind(&GraphicalAudio::find_element_and_do_action_add, this, s)));

	render->current.add(p);
      } else
	print_dir(screen_pos, j, x, y, s);

    } else {

      PObj *p = new PObj(themes->audio_file, x-5, y-5, 3, NOSCALING);
      p->x = p->real_x = x + (image_width_all - (p->w + 10))/2;
      p->y = p->real_y = y + (image_height_all_eks_text - (p->h + 10))/2;
      render->current.add(p);

      if (!s.is_fetched() && s.type != "web" && s.type != "media-file" && s.type != "media-track") {
	Dbaudiofile tmp_dbaf(s.path);
	tmp_dbaf.set_values(s);
	vector_lookup(cur_files, start) = tmp_dbaf;
	s = tmp_dbaf;
      }

      S_Touch::get_instance()->register_area(TouchArea(rect(p->x, p->y, p->w, p->h), p->layer, boost::bind(&GraphicalAudio::find_element_and_do_action_add, this, s)));

    }

    string name = get_name_from_file(s);

    string old_name = name;

    string_format::format_to_size(name, "Vera/13", image_width_all-10, false);

    string name_second_part = old_name.substr(name.size());

    if (name_second_part.size() > 0 && name_second_part[0] != ' ') {
      // splitting words
      string::size_type last_whole_word = name.rfind(' ');
      if (last_whole_word != string::npos) {
	name = name.substr(0, last_whole_word);
	name_second_part = old_name.substr(name.size());
      }
    }

    int x_size = string_format::calculate_string_width(name, "Vera/13");

    string_format::format_to_size(name_second_part, "Vera/13", image_width_all-10, false);

    int x_size2 = string_format::calculate_string_width(name_second_part, "Vera/13");

    if (screen_pos == j) {
      render->current.add(new TObj(name, "Vera/13", x + (image_width_all-x_size)/2,
				   y+image_height_all_eks_text,
				   themes->audio_marked_font1, themes->audio_marked_font2,
				   themes->audio_marked_font3, 3));
      render->current.add(new TObj(name_second_part, "Vera/13", x + (image_width_all-x_size2)/2,
				   y+image_height_all_eks_text+20,
				   themes->audio_marked_font1, themes->audio_marked_font2,
				   themes->audio_marked_font3, 3));
    } else {
      render->current.add(new TObj(name, "Vera/13", x + (image_width_all-x_size)/2,
				   y+image_height_all_eks_text,
				   themes->audio_font1, themes->audio_font2,
				   themes->audio_font3, 3));
      render->current.add(new TObj(name_second_part, "Vera/13", x + (image_width_all-x_size2)/2,
				   y+image_height_all_eks_text+20,
				   themes->audio_font1, themes->audio_font2,
				   themes->audio_font3, 3));
    }

    x += image_width_all;

    if (start == cur_files.size()-1) {
      if ((j+1)%images_per_row != 0)
	j += (images_per_row-(j+1)%images_per_row);

      x = x_start;
      y += cur_image_height_all;

      start = 0;
      continue;

    } else
      ++start;

    if ((j+1)%images_per_row == 0) {
      y += cur_image_height_all;
      x = x_start;
    }
  }

  std::ostringstream position;

  if (search_mode)
    if (cur_files.size() > 0)
      position << pos_in_list + 1 << "/" << cur_files.size();
    else
      position << "";
  else
    position << folders.top().second+1 << "/" << cur_files.size();

  int x_size = string_format::calculate_string_width(position.str(), "Vera/18");

  render->current.add(new TObj(position.str(), "Vera/18", conf->p_h_res()-(60+x_size), 20,
			       themes->audio_font1, themes->audio_font2,
			       themes->audio_font3, 3));

  render->draw_and_release("audio-graphical");
}

void GraphicalAudio::print_tracklist(vector<Dbaudiofile>& cur_files)
{
  render->prepare_new_image();

  render->current.add(new PObj(themes->audio_background, 0, 0, 0, SCALE_FULL));

  PObj *back = new PObj(themes->startmenu_music_dir, 25, 10, 2, NOSCALING);

  render->current.add(back);

  if (mode == ADD)
    S_Touch::get_instance()->register_area(TouchArea(rect(back->x, back->y, back->w, back->h), back->layer, boost::bind(&GraphicalAudio::one_up, this)));

  if (themes->show_header) {

    string header = dgettext("mms-audio", "Audio Library");

    if (mode == PLAY) {

      header = dgettext("mms-audio", "Audio - Playlist");

    } else if (folders.size() > 1) {

      string top_folder = folders.top().first.front();

      if (top_folder[top_folder.size()-1] == '/')
	top_folder = top_folder.substr(0, top_folder.size()-1);

      assert(top_folder.rfind('/') != string::npos);
      header += " - " + top_folder.substr(top_folder.rfind('/')+1);
      string_format::format_to_size(header, "Vera/28", conf->p_h_res()-220, false);
    }

    TObj *t = new TObj(header, "Vera/28", 100, 5, themes->audio_header_font1, themes->audio_header_font2,
		       themes->audio_header_font3, 2);

    render->current.add(t);

    S_Touch::get_instance()->register_area(TouchArea(rect(t->x, t->y, t->w, t->h), t->layer, boost::bind(&GraphicalAudio::toggle_playlist, this)));
  }

  if (search_mode) {

    int y = 60;
    if (conf->p_h_res() == 720)
      y -= 3;

    if (offset == -1)
      render->current.add(new RObj(0, y, conf->p_h_res(), 45, 0, 0, 0, 215, 2));
    else
      render->current.add(new RObj(0, y, conf->p_h_res(), 45,
				   themes->general_search_rectangle_color1,
				   themes->general_search_rectangle_color2,
				   themes->general_search_rectangle_color3,
				   themes->general_search_rectangle_color4, 2));

    render->current.add(new PObj(themes->general_search, 47, y, 3, NOSCALING));

    int max_x = conf->p_h_res()-250;

    string sw = search_str;

    string s = dgettext("mms-audio", "Search: ");
    int x_size = string_format::calculate_string_width(s, "Vera/22");
    string_format::format_to_size(sw, "Vera/22", max_x-x_size, true);

    render->current.add(new TObj(s + sw, "Vera/22", 100, y+5,
				 themes->search_font1, themes->search_font2,
				 themes->search_font3, 3));

    if (mode != PLAY)
      render->current.add(new TObj(search_depth, "Vera/17", 50+max_x+25, y+10, themes->search_font1,
				   themes->search_font2, themes->search_font3, 3));
  }

  if (search_mode && conf->p_input() != "keyboard")
    search_top_size += 30;

  string pic = "";
  if (position_int() < cur_files.size())
    pic = get_cover(vector_lookup(*files, position_int()));

  int pos = 0;

#ifdef use_mouse

  if (audio_conf->p_display_buttons() && mode == PLAY) {

    int y_buttons = (conf->p_v_res()-7*40)/2;

    PObj *p = new PObj("buttons/PLAY.png", 20, y_buttons+0*40, 2, NOSCALING);
    render->current.add(p);
    S_Touch::get_instance()->register_area(TouchArea(rect(p->x, p->y, p->w, p->h), p->layer, boost::bind(&GraphicalAudio::play, this)));

    p = new PObj("buttons/PAUSE.png", 20, y_buttons+1*40, 2, NOSCALING);
    render->current.add(p);
    S_Touch::get_instance()->register_area(TouchArea(rect(p->x, p->y, p->w, p->h), p->layer, boost::bind(&GraphicalAudio::pause, this)));

    p = new PObj("buttons/STOP.png", 20, y_buttons+2*40, 2, NOSCALING);
    render->current.add(p);
    S_Touch::get_instance()->register_area(TouchArea(rect(p->x, p->y, p->w, p->h), p->layer, boost::bind(&GraphicalAudio::stop, this)));

    p = new PObj("buttons/PREV.png", 20, y_buttons+3*40, 2, NOSCALING);
    render->current.add(p);
    S_Touch::get_instance()->register_area(TouchArea(rect(p->x, p->y, p->w, p->h), p->layer, boost::bind(&GraphicalAudio::prev_track, this)));

    p = new PObj("buttons/NEXT.png", 20, y_buttons+4*40, 2, NOSCALING);
    render->current.add(p);
    S_Touch::get_instance()->register_area(TouchArea(rect(p->x, p->y, p->w, p->h), p->layer, boost::bind(&GraphicalAudio::next_track, this)));

    p = new PObj("buttons/REV.png", 20, y_buttons+5*40, 2, NOSCALING);
    render->current.add(p);
    S_Touch::get_instance()->register_area(TouchArea(rect(p->x, p->y, p->w, p->h), p->layer, boost::bind(&GraphicalAudio::fb, this)));

    p = new PObj("buttons/FF.png", 20, y_buttons+6*40, 2, NOSCALING);
    render->current.add(p);
    S_Touch::get_instance()->register_area(TouchArea(rect(p->x, p->y, p->w, p->h), p->layer, boost::bind(&GraphicalAudio::ff, this)));
  }
#endif

  if (cur_files.size() > 0) {

    pos = position_int();
    if (search_mode)
      pos = offset;

    Dbaudiofile currently_selected;
    if (pos != -1)
      currently_selected = vector_lookup(cur_files, pos);
    else
      pos = 0;

    print_range_modifying<Dbaudiofile>(cur_files, currently_selected, pos,
				       boost::bind(&GraphicalAudio::print_element, this, _1, _2, _3, !pic.empty()), 28);

    if (mode == ADD && (search_depth == dgettext("mms-audio", "current folder") || !search_mode)) {

      if (!pic.empty()) {

	if (cur_files.size() > 0) {

	  Dbaudiofile dbaf = vector_lookup(cur_files, pos);

	  int max_x_pic = 240, max_y_pic = 266;

	  if (conf->p_h_res() == 1024) {
	    max_x_pic = 300;
	    max_y_pic = 333;
	  } else if (conf->p_v_res() == 480) {
	    max_x_pic = 200;
	    max_y_pic = 222;
	  } else if (conf->p_v_res() == 405) {
	    max_x_pic = 150;
	    max_y_pic = 167;
	  }

	  string artist = string_format::convert(string_format::trim(dbaf.artist), true);
	  string album = string_format::convert(string_format::trim(dbaf.album), true);

	  string_format::format_to_size(artist, "Vera/20", max_x_pic, false);
	  string_format::format_to_size(album, "Vera/18", max_x_pic, false);

	  int y_artist = 100;
	  int y_album = 135;
	  int y_image = 183;

	  if (search_mode) {
	    y_artist += search_top_size;
	    y_album += search_top_size;
	    y_image += search_top_size;
	  }

	  int x_start = 358;

	  if (conf->p_h_res() == 1024)
	    x_start = 418;

	  if (conf->p_v_res() == 405)
	    x_start = 250;

	  x_start = conf->p_h_res()-x_start+25;

	  render->current.add(new TObj(artist, "Vera/20", x_start, y_artist,
				       themes->audio_font1, themes->audio_font2,
				       themes->audio_font3, 3));

	  render->current.add(new TObj(album, "Vera/18", x_start, y_album,
				       themes->audio_font1, themes->audio_font2,
				       themes->audio_font3, 3));

	  // scale it
	  render->create_scaled_image_wrapper(pic, max_x_pic, max_y_pic);
	  render->current.add(new PFObj(pic, x_start, y_image, max_x_pic, max_y_pic, true, 3));
	}
      }
    }
  }

  print_buttom(cur_files.size(), pos);

  if (search_mode && conf->p_input() != "keyboard")
    search_top_size -= 30;
}

vector<Dbaudiofile> GraphicalAudio::add_dir(const string& dir, bool from_media)
{
  int nr_files = 0;
  for (file_iterator<file_t, default_order> i (dir); i != i.end (); ++i)
    if (!isDirectory(i->getName()) && check_type(i->getName(), audio_conf->p_filetypes_a()) != emptyMyPair)
      ++nr_files;

  ScreenUpdater *screen_updater = S_ScreenUpdater::get_instance();

  screen_updater->timer.deactivate("audio_fullscreen");

  screen_updater->status_progressbar = 0;
  screen_updater->total_progressbar = nr_files;
  screen_updater->header = dgettext("mms-audio", "Extracting metadata");

  screen_updater->timer.add(TimeElement("progressbar", &check_progressbar, &print_progressbar));

  vector<Dbaudiofile> cur_files;

  string cur_directory = "";
  string parent_dir = dir;
  vector<Dbaudiofile> files_in_dir;

  libfs_set_folders_first(opts.dir_order());

  for (file_iterator<file_t, default_order> i (dir); i != i.end (); ++i)
    {
      if (global->check_stop_bit()) {
	cur_files.clear();
	break;
      }

      string filename = filesystem::FExpand(i->getName());

      if (cur_directory != i.currentDirectory()) {
	cur_directory = i.currentDirectory();
	add_tracks_from_dir(files_in_dir, cur_files);
      }

      if (isDirectory(filename)) {
	insert_file_into_db(filename, cur_directory);
      } else {
	const MyPair filetype = check_type(filename, audio_conf->p_filetypes_a());

	if (filetype != emptyMyPair) {

	  insert_file_into_db(filename, cur_directory);

	  Dbaudiofile d(addsimplefile(filename, filetype, from_media));
	  d.db_id = db_id(filename, false);
	  files_in_dir.push_back(d);

	  ++screen_updater->status_progressbar;
	}
      }
    }

  // add tracks from the last directory
  add_tracks_from_dir(files_in_dir, cur_files);

  screen_updater->timer.del("progressbar");
  cleanup_progressbar();

  screen_updater->timer.activate("audio_fullscreen");

  return cur_files;
}

void GraphicalAudio::remove_track_from_playlist(const Dbaudiofile& p)
{
  assert(p.playlist_id != -1);

  db_mutex.enterMutex();

  char *tmp = sqlite3_mprintf("DELETE FROM Playlist WHERE id='%q'", conv::itos(p.playlist_id).c_str());
  db.execute(tmp);
  sqlite3_free(tmp);

  db_mutex.leaveMutex();
}

void GraphicalAudio::remove_all_tracks_from_playlist()
{
  db_mutex.enterMutex();
  db.execute("DELETE FROM Playlist");
  db_mutex.leaveMutex();
}

void GraphicalAudio::ids(Dbaudiofile& p)
{
  p.db_id = db_id(p.path, false);
}

void GraphicalAudio::add_track_to_playlist(Dbaudiofile p)
{
  db_mutex.enterMutex();

  char *tmp = sqlite3_mprintf("INSERT INTO Playlist VALUES(NULL, '%q')", conv::itos(p.db_id).c_str());
  db.execute(tmp);
  sqlite3_free(tmp);

  p.playlist_id = db.last_index();

  db_mutex.leaveMutex();

  AudioTemplate<Dbaudiofile>::add_track_to_playlist(p);
}

void GraphicalAudio::add_tracks_to_playlist(const std::vector<Dbaudiofile>& vp)
{
  db_mutex.enterMutex();

  db.execute("BEGIN TRANSACTION");

  foreach (Dbaudiofile p, vp) {
    char *tmp = sqlite3_mprintf("INSERT INTO Playlist VALUES(NULL, '%q')", conv::itos(p.db_id).c_str());
    db.execute(tmp);
    sqlite3_free(tmp);

    p.playlist_id = db.last_index();

    AudioTemplate<Dbaudiofile>::add_track_to_playlist(p);
  }

  db.execute("END TRANSACTION");

  db_mutex.leaveMutex();

  save_playlist("last");
}

void GraphicalAudio::add_tracks_from_dir(vector<Dbaudiofile> &files_in_dir,
					 vector<Dbaudiofile> &cur_files)
{
  // check if all the files are missing the track tag
  bool missing_track = true;
  foreach (Dbaudiofile& file, files_in_dir)
    if (file.track != 0) {
      missing_track = false;
      break;
    }

  if (missing_track) {

    // just add all the tracks, we don't care about the order since its
    // probably ok already
    foreach (Dbaudiofile& file, files_in_dir)
      cur_files.push_back(file);

    files_in_dir.clear();

  } else {

    while (!files_in_dir.empty()) {
      // find the track with the smallest id
      int smallest_track = 1000000;
      vector<Dbaudiofile>::iterator element;

      for (vector<Dbaudiofile>::iterator fid = files_in_dir.begin(),
	     fid_end = files_in_dir.end(); fid != fid_end; ++fid)
	if (fid->track < smallest_track) {
	  smallest_track = fid->track;
	  element = fid;
	}

      assert(smallest_track != 1000000);
      cur_files.push_back(*element);
      files_in_dir.erase(element);
    }
  }
}

bool GraphicalAudio::rdir_internal(const string& filename, const string& argv,
				   vector<Dbaudiofile> &cur_files, bool in_db)
{
  if (global->check_stop_bit()) {
    cur_files.clear();
    return false;
  }

  if (isDirectory (filename))
    {
      // dir
      Simplefile r;
      r.id = ++id;
      r.name = filename.substr(argv[argv.size()-1] == '/' ? argv.size() : argv.size() + 1);

      if (r.name[r.name.size()-1] == '/')
	r.name = r.name.substr(0, r.name.size()-1);

      if (conf->p_convert())
	r.name = string_format::convert(r.name);
      r.lowercase_name = string_format::lowercase(r.name);
      if (filename[filename.size()-1] != '/')
	r.path = filename + '/';
      else
	r.path = filename;
      r.type = "dir";
      Dbaudiofile d(r);
      if (in_db) {
	d.db_id = db_id(filename, true);
	assert(d.db_id != -1);
      }
      cur_files.push_back(d);
    } else {
      // file
      const MyPair filetype = check_type(filename, audio_conf->p_filetypes_a());
      if (filetype != emptyMyPair) {
	Dbaudiofile d(addsimplefile(filename, filetype, navigating_media));
	if (in_db) {
	  d.db_id = db_id(filename, false);
	  assert(d.db_id != -1);
	}
	cur_files.push_back(d);

	++(S_ScreenUpdater::get_instance())->status_progressbar;
      }
    }

  return true;
}

vector<string> GraphicalAudio::check_db_for_folders(const string& parent)
{
  vector<string> names;

  string good_parent = string_format::unique_folder_name(parent);

  db_mutex.enterMutex();
  SQLQuery *q = db.query("Folders", ("SELECT id FROM %t WHERE filename='" + string_format::escape_db_string(good_parent) + "'").c_str());
  if (q && q->numberOfTuples() > 0) {
    assert(q->numberOfTuples() == 1);
    SQLRow &row = (*q)[0];
    string parent_id = row["id"];

    SQLQuery *q2;

    if (opts.dir_order() == dgettext("mms-audio", "directories first"))
      q2 = db.query("Folders", ("SELECT filename, upper(name) as key FROM %t WHERE parent='" + parent_id + "' ORDER BY is_folder desc, key").c_str());
    else
      q2 = db.query("Folders", ("SELECT filename, upper(name) as key FROM %t WHERE parent='" + parent_id + "' ORDER BY key").c_str());

    if (q2) {
      for (int i = 0; i < q2->numberOfTuples(); ++i) {
	SQLRow &row = (*q2)[i];
	names.push_back(row["filename"]);
      }
    }
    delete q2;
  }
  delete q;
  db_mutex.leaveMutex();

  return names;
}

void GraphicalAudio::insert_file_into_db(const string& filename, const string& parent)
{
  bool is_dir = isDirectory(filename);

  string good_filename;

  if (is_dir)
    good_filename = string_format::unique_folder_name(filename);
  else
    good_filename = filename;

  db_mutex.enterMutex();

  // check if already exists
  SQLQuery *q = db.query("Folders", ("SELECT id FROM %t WHERE filename='" + string_format::escape_db_string(good_filename) + "'").c_str());
  if (q && q->numberOfTuples() > 0) {
    delete q;
    db_mutex.leaveMutex();
    return;
  } else
    delete q;

  string parent_id = get_parent_id(parent, db, db_mutex, navigating_media ? top_media_folders : audio_folders);

  string::size_type i;
  if ((i = filename.rfind('/')) == string::npos)
    i = 0;
  string name = filename.substr((i != 0) ? i+1 : i);
  string path = filename.substr(0, i+1);

  if (!is_dir && (i = name.rfind('.')) != string::npos)
    name = name.substr(0, i);

  db_mutex.enterMutex();
  char *tmp_insert = sqlite3_mprintf("INSERT INTO Folders VALUES(NULL, '%q', '%q', '%q', '%q', '%q')",
				     parent_id.c_str(), good_filename.c_str(), name.c_str(),
				     string_format::lowercase(name).c_str(), conv::itos(is_dir).c_str());
  db.execute(tmp_insert);
  sqlite3_free(tmp_insert);
  db_mutex.leaveMutex();

  // and cover

  if (is_dir)
    path = good_filename;

  insert_cover_into_db(path, is_dir ? "dir" : "file");
}

void GraphicalAudio::insert_cover_into_db(const string& filename, const string& type)
{
  bool found_cover = false;

  db_mutex.enterMutex();

  SQLQuery *q2 = db.query("Covers", ("SELECT Cover FROM %t WHERE Path='" + string_format::escape_db_string(filename) + "'").c_str());
  if (q2 && q2->numberOfTuples() > 0) {
    found_cover = true;
  }
  delete q2;

  db_mutex.leaveMutex();

  if (!found_cover) {

    string cover = "";

    if (type == "file") {
      GSimplefile s;
      s.path = filename;
      cover = s.find_cover_in_current_dir();
    } else if (type == "media-track") {
      recurse_mkdir(conf->p_var_data_dir(),"cddb_data", &cover);
      cover += filename;
      cover += ".jpg";

      // check if the cover even exists before inserting it
      if (!file_exists(cover))
	cover = "";

    } else if (type == "dir") {
      cover = get_first_cover_in_dir(filename);
    } else {
      assert(false);
    }

    db_mutex.enterMutex();

    char *tmp_insert_2 = sqlite3_mprintf("INSERT INTO Covers VALUES('%q', '%q')",
					 filename.c_str(), cover.c_str());
    db.execute(tmp_insert_2);
    sqlite3_free(tmp_insert_2);

    db_mutex.leaveMutex();
  }
}

vector<Dbaudiofile> GraphicalAudio::rdir_hd(const string& argv, bool insert_into_db)
{
  libfs_set_folders_first(opts.dir_order());

  if (insert_into_db) {

    bool inserted = false;

    for (file_iterator<file_t, default_order> i (argv); i != i.end (); i.advance(false)) {
      string filename = filesystem::FExpand(i->getName());
      if (isDirectory(filename) || check_type(filename, audio_conf->p_filetypes_a()) != emptyMyPair) {
	insert_file_into_db(filename, argv);
	inserted = true;
      }
    }

    if (inserted)
      return rdir(argv);
    else {
      vector<Dbaudiofile> cur_files;
      return cur_files;
    }

  } else {

    vector<Dbaudiofile> cur_files;

    for (file_iterator<file_t, default_order> i (argv); i != i.end (); i.advance(false)) {
      string filename = filesystem::FExpand(i->getName());
      if (!rdir_internal(filename, argv, cur_files, false))
	break;
    }

    return cur_files;
  }
}

vector<Dbaudiofile> GraphicalAudio::rdir(const string& argv)
{
  vector<Dbaudiofile> cur_files;

  vector<string> db_files = check_db_for_folders(argv);

  if (db_files.size() > 0) {
    foreach (string& file, db_files)
      if (!rdir_internal(file, argv, cur_files, true))
	break;
  } else {
    cur_files = rdir_hd(argv, true);
  }

  vector<Dbaudiofile> sorted_by_id_files;

  add_tracks_from_dir(cur_files, sorted_by_id_files);

  return sorted_by_id_files;
}

void GraphicalAudio::read_dirs()
{
  while (!all_metadata_files.empty())
    all_metadata_files.pop();

  vector<Dbaudiofile> playlist_new;

  int nr_files = 0;

  foreach (string& folder, audio_folders)
    for (file_iterator<file_t, default_order> i (folder); i != i.end (); i.advance(false))
      if (!isDirectory (i->getName()) &&
	  check_type(i->getName(), audio_conf->p_filetypes_a()) != emptyMyPair)
	++nr_files;

  ScreenUpdater *screen_updater = S_ScreenUpdater::get_instance();

  screen_updater->status_progressbar = 0;
  screen_updater->total_progressbar = nr_files;
  screen_updater->header = dgettext("mms-audio", "Extracting metadata");

  screen_updater->timer.add(TimeElement("progressbar", &check_progressbar, &print_progressbar));

  // process audio dir
  audio_list = parse_dir(audio_folders);

  screen_updater->timer.del("progressbar");
  cleanup_progressbar();

  if (audio_folders.size() > 1)
    std::sort(audio_list.begin(), audio_list.end(), file_sort());

  if (audio_list.size() != 0)
    // set position to the first file
    folders.top().second = 0;

  if (mode == ADD)
    files = &audio_list;
  else
    files = &playlist;

  S_BackgroundUpdater::get_instance()->run_once(boost::bind(&GraphicalAudio::check_for_changes, this));
}

void GraphicalAudio::intelligent_random_next()
{
#ifdef use_imms
  db_mutex.enterMutex();

  imms->playlist_changed(playlist.size());
  next_track_pos = -1;
  //  imms->select_next();

  int time_left = audio_state->p->p_total_time() - audio_state->p->p_cur_time();

  if (time_left < 3)
    time_left = 0;

  bool forced = time_left;

  Simplefile cur_nr = audio_state->p->p_cur_nr();

  if (imms_cur_path != cur_nr.path) {
    // something strange happened, we are not playing the track we IMMS thinks
    // we are
    for (int i = 0, end = playlist.size(); i < end; ++i)
      if (vector_lookup(playlist, i).id == cur_nr.id) {
	imms->start_song(i, cur_nr.path);
	break;
      }
  }

  imms->end_song(!time_left, forced, false);

  imms->select_next();

  int i = 0;

  while (next_track_pos == -1) {
    if (i > 20) {
      std::cerr << dgettext("mms-audio", "something wrong in intelligent random") << std::endl;
      db_mutex.leaveMutex();
      return;
    }

    usleep(50*1000); // wait for next track
    ++i;
  }

  audio_state->p->set_cur_nr(vector_lookup(playlist, next_track_pos));

  imms_cur_path = audio_state->p->p_cur_nr().path;

  imms->start_song(next_track_pos, imms_cur_path);

  audio_state->direction = Audio_s::ADDED;

  db_mutex.leaveMutex();
#endif
}

void GraphicalAudio::intelligent_random_prev()
{
#ifdef use_imms
  intelligent_random_next();
#endif
}

void GraphicalAudio::extract_metadata()
{
  if (!extracted_all_files && (conf->p_idle_time() > 0) &&
      (opts.metadata_extract() == dgettext("mms-audio", "yes"))) {

    int nr_extracts = 3;

    // check if we are idle
    int idle_time = time(0)-conf->p_last_key();
    if (idle_time < 3)
      return;
    else if (idle_time > 60 * conf->p_idle_time())
      nr_extracts = 4;

    if (all_metadata_files.empty()) {
      list<string> dirs = audio_folders;
      libfs_set_folders_first(opts.dir_order());
      foreach (string& dir, dirs)
	for (file_iterator<file_t, default_order> i (dir); i != i.end (); ++i)
	  all_metadata_files.push(i->getName());
      // optimization, we don't want to both make the list and start extracting
      return;
    }

    while (!all_metadata_files.empty()) {
      if (!file_exists_in_db(all_metadata_files.top())) {
	Dbaudiofile(all_metadata_files.top());
	if (nr_extracts == 0)
	  return;
	else
	  --nr_extracts;
      }
      all_metadata_files.pop();
    }

    extracted_all_files = true;
  }
}

void GraphicalAudio::startup_updater()
{
  Audio::startup_updater();

  S_BackgroundUpdater::get_instance()->timer.add(TimeElement("search marker", boost::bind(&GraphicalAudio::check_metadata,
											  this),
							     boost::bind(&GraphicalAudio::extract_metadata, this)));

  S_Search::get_instance()->register_module(SearchModule(dgettext("mms-audio", "Audio"),
							 boost::bind(&GraphicalAudio::gen_search_list, this, _1),
							 boost::bind(&GraphicalAudio::reenter, this,  _1)));
}
