#define _XOPEN_SOURCE /* glibc2 needs this */

#include "config.h"

#include "epg.hpp"

#include "input.hpp"
#include "common.hpp"
#include "graphics.hpp"
#include "updater.hpp"
#include "common-feature.hpp"
#include "startmenu.hpp"
#include "busy_indicator.hpp"
#include "shutdown.hpp"
#include "extra_menu.hpp"
#include "touch.hpp"

#include "search.hpp"

#include "plugins.hpp"

#include <time.h>

#include "gettext.hpp"

#include "boost.hpp"

#include <iostream>
#include <list>
#include <sstream>
#include <algorithm>

// time of last change
#include <sys/stat.h>

#include "tinyxml.h"

#ifdef use_nls
#include <langinfo.h>
#include <locale.h>
#endif

using std::string;
using std::vector;
using std::list;

pthread_mutex_t Epg::singleton_mutex = PTHREAD_MUTEX_INITIALIZER;

time_t Programme::str_to_time(const char *s) {
  struct tm tm;
  char * datetimetz;

#ifdef __UCLIBC__
  char Y[5], m[3], d[3], H[3], M[3];
  char t[32];

  s += 0; strncpy(Y, s, 4); Y[4] = 0;
  s += 4; strncpy(m, s, 2); m[2] = 0;
  s += 2; strncpy(d, s, 2); d[2] = 0;
  s += 2; strncpy(H, s, 2); H[2] = 0;
  s += 2; strncpy(M, s, 2); M[2] = 0;
  sprintf(t, "%s-%s-%s %s:%s:00 +0000", Y, m, d, H, M);
#endif

  int hours=0;
  int minutes=0;
  int factor=1;

  //fix unset value
  tm.tm_isdst=0;
  //get timezone
  tzset();

#ifdef __UCLIBC__
  datetimetz = strptime(t, "%Y-%m-%d %H:%M:%S", &tm);
#else
  datetimetz = strptime(s, "%Y%m%d%H%M%S", &tm);
#endif

  //check if data reliable
  if (datetimetz==NULL || strlen(datetimetz)!=6)
    return-1;
  //move up space
  datetimetz++;
  //get factor
  if (datetimetz[0]=='-') {
    factor=-1;
  }
  //move up sign
  datetimetz++;
  //copy XXXX to extract hours
  char tmptz[5];
  strcpy(tmptz,datetimetz);
  //we only need hours
  tmptz[2]=0;
  //get hours
  hours = atoi(tmptz);
  //get minutes
  minutes= atoi(&datetimetz[2]);
  //correction using local timezone and xml timezone to obtain the
  //right date time
  tm.tm_sec+=timezone + factor*((minutes * 60)+(hours * 3600));
  return mktime(&tm);
}

bool Programme::operator<(const Programme& rhs) const
{
  return (this->start < rhs.start);
}

Programme::Programme(TiXmlElement *element)
  : start(0), stop(0)
{
  TiXmlNode *node;
  TiXmlText *text;

  if (element->Attribute("start")) {
    start = str_to_time(element->Attribute("start"));
  }

  if (element->Attribute("stop")) {
    stop = str_to_time(element->Attribute("stop"));
  }

  if (element->FirstChild("title")) {
    node = element->FirstChild("title");
    if (node) {
      node = node->FirstChild();
      if (node) {
	text = node->ToText();
	if (text)
	  title = text->Value();
      }
    }
  }

  if (element->FirstChild("desc")) {
    node = element->FirstChild("desc");
    if (node) {
      node = node->FirstChild();
      if (node) {
	text = node->ToText();
	if (text)
	  desc = text->Value();
      }
    }
  }
}

Channel::Channel(TiXmlElement *element)
{
  TiXmlNode *node = 0;
  TiXmlText *text = 0;

  if (element->Attribute("id")) {
    id = element->Attribute("id");
  }

  if (element->FirstChild("display-name")) {
    node = element->FirstChild("display-name");
    if (node != 0) {
      node = node->FirstChild();
      if (node != 0) {
	text = node->ToText();
	display_name = text->Value();
      }
    }
  }
}

void Epg::create_db()
{
  vector<string> fields;
  fields.push_back("version INTEGER NOT NULL");
  fields.push_back("date TEXT");
  db.createTable("Schema", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("tvid TEXT");
  fields.push_back("name TEXT");
  db.createTable("Channel", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("cid TEXT"); // channel id
  fields.push_back("title TEXT");
  fields.push_back("ltitle TEXT");
  fields.push_back("description TEXT");
  fields.push_back("ldescription TEXT");
  fields.push_back("start_time TEXT");
  fields.push_back("end_time TEXT");
  db.createTable("Program", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("pid INTEGER NOT NULL");
  fields.push_back("year INTEGER NOT NULL");
  db.createTable("MovieProgram", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("pid INTEGER NOT NULL");
  fields.push_back("nr INTEGER NOT NULL");
  db.createTable("EpisodeProgram", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("name TEXT");
  fields.push_back("role TEXT");
  db.createTable("Actors", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("aid INTEGER NOT NULL");
  fields.push_back("pid INTEGER NOT NULL");
  db.createTable("AProgram", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("name TEXT");
  db.createTable("Directors", fields);

  fields.clear();
  fields.push_back("id INTEGER PRIMARY KEY");
  fields.push_back("did INTEGER NOT NULL");
  fields.push_back("pid INTEGER NOT NULL");
  db.createTable("DProgram", fields);

  create_indexes();
}

void Epg::create_indexes()
{
  db.execute("CREATE INDEX idx_program_title ON Program (ltitle)");
  db.execute("CREATE INDEX idx_program_desc ON Program (ldescription)");
  db.execute("CREATE INDEX idx_channel_tvid ON Channel (tvid)");
}

void Epg::clear_db()
{
  SQLQuery *dq = db.query("sqlite_master", "SELECT name FROM %t WHERE type='table'");
  if (dq) {
    for (int i = 0; i < dq->numberOfTuples(); ++i) {
      db.execute(("DROP TABLE " + (*dq)[i]["name"]).c_str());
    }
  }

  delete dq;

  dq = db.query("sqlite_master", "SELECT name FROM %t WHERE type='index'");
  if (dq) {
    for (int i = 0; i < dq->numberOfTuples(); ++i) {
      db.execute(("DROP INDEX " + (*dq)[i]["name"]).c_str());
    }
  }

  delete dq;
}

void Epg::insert_schema()
{
  struct stat s;
  stat(epg_conf->p_epg_data().c_str(), &s);

  char *tmp = sqlite3_mprintf("INSERT INTO Schema VALUES('3', '%q')",
			      conv::itos(s.st_mtime).c_str());
  db.execute(tmp);
  sqlite3_free(tmp);
}

bool Epg::check_db()
{
  if (!file_exists(epg_conf->p_epg_data()))
    return true;

  bool db_ok = true;

  if (db.hasTable("Schema")) {
    // check for correct version
    SQLQuery *q = db.query("Schema", "SELECT * FROM %t");
    if (q && q->numberOfTuples() > 0) {
      SQLRow &row = (*q)[0];
      struct stat s;
      stat(epg_conf->p_epg_data().c_str(), &s);
      if (row["date"] != conv::itos(s.st_mtime)) {
	clear_db();
	create_db();
	insert_schema();
	db_ok = false;
      } else if (row["version"] != "3") {

	Print print(gettext("Upgrading EPG database schema"), Print::INFO);

	clear_db();
	create_db();
	insert_schema();
      }
    } else {
      clear_db();
      create_db();
      insert_schema();
      db_ok = false;
    }
    delete q;
  } else {
    // create the whole db
    clear_db();
    create_db();
    insert_schema();
    db_ok = false;
  }

  return db_ok;
}

bool Epg::read_xml_file(list<Channel> *chan, bool db_ok)
{
  if (!file_exists(epg_conf->p_epg_data()))
    return false;

  db_mutex.enterMutex();

  if (!db_ok) {
    DialogWaitPrint pdialog(dgettext("mms-epg", "Inserting EPG data into database"), 2000);

    db.execute("BEGIN TRANSACTION");
  }

  TiXmlDocument doc(epg_conf->p_epg_data().c_str());
  TiXmlElement* root_element;
  TiXmlElement* element;

  doc.LoadFile();
  root_element = doc.RootElement();

  if (root_element == 0) {
    db_mutex.leaveMutex();
    return false;
  }

  for (element = root_element->FirstChildElement("channel"); element != 0;
       element = element->NextSiblingElement("channel")) {
    Channel temp = Channel(element);
    chan->push_back(temp);
    if (!db_ok) {
      char *tmp = sqlite3_mprintf("INSERT INTO Channel VALUES(NULL, '%q', '%q')",
				  temp.id.c_str(), temp.display_name.c_str());
      db.execute(tmp);
      sqlite3_free(tmp);
    }
  }

  list<Channel>::iterator channel_iterator = chan->begin();

  string last_id = "";

  for (element = root_element->FirstChildElement("programme"); element != 0;
       element = element->NextSiblingElement("programme")) {

    if (last_id != element->Attribute("channel")) // find a new channel
      for (channel_iterator = chan->begin();
	   channel_iterator != chan->end(); ++channel_iterator)
	if (channel_iterator->id == element->Attribute("channel")) {
	  last_id = channel_iterator->id;
	  break;
	}

    Programme p = Programme(element);

    if (channel_iterator->programme.size() > 0)
      if (channel_iterator->programme.back().stop == 0)
	channel_iterator->programme.back().stop = p.start;

    channel_iterator->programme.push_back(p);

    if (!db_ok) {
      // FIXME: determine if movie, episode or ordinary program

      char *tmp = sqlite3_mprintf("INSERT INTO Program VALUES(NULL, '%q', '%q', '%q', '%q', '%q', '%q', '%q')",
				  channel_iterator->id.c_str(), p.title.c_str(),
				  string_format::lowercase(p.title).c_str(),
				  p.desc.c_str(), string_format::lowercase(p.desc).c_str(),
				  conv::itos(p.start).c_str(), conv::itos(p.stop).c_str());
      db.execute(tmp);
      sqlite3_free(tmp);
    }
  }

  for (list<Channel>::iterator chan_iter = chan->begin(), end = chan->end();
       chan_iter != end;) {
    if (chan_iter->programme.size() > 0) {
      std::sort(chan_iter->programme.begin(), chan_iter->programme.end());
      ++chan_iter;
    } else
      chan_iter = chan->erase(chan_iter);
  }

  if (!db_ok)
    db.execute("END TRANSACTION");

  db_mutex.leaveMutex();

  return true;
}

Epg::EpgOpts::EpgOpts()
{
#ifdef use_nls
  // gettext
  setlocale(LC_ALL, ""); /* set from the environment variables */
  bindtextdomain("mms-epg", mms_prefix"/share/locale");
  bind_textdomain_codeset("mms-epg", nl_langinfo(CODESET));
#endif

  // header
  header = "EPG Options";
  translated_header = dgettext("mms-epg", "EPG Options");
  save_name = "EPGOptions";

  // values

  // reload
  vector<string> none;
  reload_p = new Option(false, dgettext("mms-epg", "reload"), "reload", 0, none, none);
  val.push_back(reload_p);
}

Epg::EpgOpts::~EpgOpts()
{
  save();

  delete reload_p;
}

void Epg::update_time()
{
  struct tm *tm;

  cur_time = time(0);
  tm = localtime(&cur_time);
  if (tm->tm_min < 30) {
    tm->tm_min = 0;
  } else {
    tm->tm_min = 30;
  }
  tm->tm_sec = 0;
  cur_time = mktime(tm);
  sel_time = cur_time;
}

void Epg::read_db()
{
  if (!file_exists(epg_conf->p_epg_data()))
    return;

  channels.clear();

  if (!read_xml_file(&channels, check_db()))
    return;

  cur_channel = channels.begin();
  sel_channel = cur_channel;
  sel_channel_num = 0;

  update_time();

  for (program_iter = sel_channel->programme.begin(),
	 program_iter_end = sel_channel->programme.end();
       program_iter != program_iter_end; ++program_iter)
    if (program_iter->stop >= sel_time)
      break;

  cur_time = sel_time = program_iter->start;
}

Epg::Epg(int x_start_value)
  : db((conf->p_homedir() + "epg.db").c_str()),
    exit_loop(false), x_start(x_start_value),
    epg_script_found(true), update_running(false)
{
  epg_conf = S_EPGConfig::get_instance();

  if (conf->p_h_res() == 720)
    y_start = 130;
  else if (conf->p_h_res() == 800)
    y_start = 140;
  else if (conf->p_h_res() == 1024)
    y_start = 170;
  else if (conf->p_h_res() == 1280)
    y_start = 170;
  else
    y_start = 100; // default value

  rows = (conf->p_v_res()-(y_start+10+90+25))/40;

  if (conf->p_v_res() == 405)
    rows = (conf->p_v_res()-(y_start+10+80+25))/40;

  timeslots = int((conf->p_h_res()-x_start-15)/(30*60/10.25));

  scale_factor = (30*60)/(double(conf->p_h_res()-x_start-25)/timeslots);

  epg_updater = new EpgUpdate();
}

Epg::~Epg()
{
  if (update_running) {
    // change our group id so that we don't ourselves when we quit (this
    // requires root privs.)
    pid_t old_gpid = getpgid(0);
    setpgid(0, old_gpid+1);

    // kill updater
    system(("kill -9 -" + conv::itos(old_gpid)).c_str());
  }

  db_mutex.enterMutex();
  db.close();
  db_mutex.leaveMutex();
}

bool Epg::valid_program()
{
  bool found_match = false;

  time_t sel_start_time = program_iter->start;

  vector<Programme>::iterator piter;
  for (piter = sel_channel->programme.begin(),
	 program_iter_end = sel_channel->programme.end();
       piter != program_iter_end; ++piter)
    if (piter->stop > sel_start_time &&
	piter->stop > cur_time) {
      found_match = true;
      break;
    }

  program_iter = piter;

  return found_match;
}

void Epg::find_next_program()
{
  vector<Programme>::iterator end = sel_channel->programme.end();
  --end;

  vector<Programme>::iterator piter;
  for (piter = sel_channel->programme.begin(),
	 program_iter_end = sel_channel->programme.end();
       piter != program_iter_end; ++piter)
    if (piter->start >= sel_time) {
      if (piter->start == sel_time && piter != end) {

	vector<Programme>::iterator last_program_iter = piter;
	++piter;

	// fix for strange programs with no length
	if (last_program_iter->start == last_program_iter->stop ||
	    piter->start == piter->stop && piter != end) {
	  ++piter;
	}
      }
      break;
    }

  program_iter = piter;

  if (program_iter == end)
    return;

  sel_time = program_iter->start;

  int time = timeslots*30*60;

  while (sel_time >= cur_time + time)
    cur_time += 30*60;
}

void Epg::find_valid_program()
{
  if (!valid_program()) {
    find_next_program();
    return;
  }

  sel_time = program_iter->start;

  int time = timeslots*30*60;

  while (sel_time >= cur_time + time)
    cur_time += 30*60;
}

bool Epg::check_tv_file()
{
  // check if the TV.xml file is new enough
  time_t max = 0;

  foreach (Channel& channel, channels)
    if (channel.programme.size() > 0) {
      vector<Programme>::iterator cha = channel.programme.end();
      --cha;
      if (cha != channel.programme.begin())
	--cha;

      max = std::max(cha->stop, max);
    }

  if (max < time(0)) {
    Print pdialog(dgettext("mms-epg", "Your TV listings file is too old or inaccessible"),
		  Print::SCREEN);

    return false;
  } else
    return true;
}

// commands
void Epg::print_description()
{
  ScreenUpdater *screen_updater = S_ScreenUpdater::get_instance();

  S_BusyIndicator::get_instance()->idle();

  if (program_iter->desc.size() != 0) {
    screen_updater->timer.deactivate("epg");
    print_description(*program_iter);
    screen_updater->timer.activate("epg");
  } else
    DialogWaitPrint pdialog(dgettext("mms-epg", "No description available"), 1500);
}

void Epg::options()
{
  ScreenUpdater *screen_updater = S_ScreenUpdater::get_instance();
  S_BusyIndicator::get_instance()->idle();

  screen_updater->timer.deactivate("epg");

  if (opts.mainloop()) {
    // reload
    read_db();
  }

  screen_updater->timer.activate("epg");
}

void Epg::exit()
{
  exit_loop = true;
}

bool Epg::move_element_up()
{
  if (sel_channel != channels.begin()) {
    --sel_channel;

    if (sel_channel_num > 0) {
      --sel_channel_num;
    } else {
      --cur_channel;
    }

    return true;
  } else
    return false;
}

bool Epg::move_element_down()
{
  if (sel_channel != (--channels.end())) {
    ++sel_channel;

    if (sel_channel_num < (rows - 1)) {
      ++sel_channel_num;
    } else {
      ++cur_channel;
    }

    return true;
  } else
    return false;
}

string Epg::mainloop()
{
  Input input;

  ScreenUpdater *screen_updater = S_ScreenUpdater::get_instance();

  // disable screensaver
  screen_updater->timer.deactivate("audio_fullscreen");

  screen_updater->timer.add(TimeElement("epg", boost::bind(&Epg::check_epg_print, this), boost::bind(&Epg::epg_print, this)));

  bool update_needed = true;

  BusyIndicator *busy_indicator = S_BusyIndicator::get_instance();
  Shutdown *sd = S_Shutdown::get_instance();

  while (!exit_loop) {

    if (update_needed)
      print();

    input = input_master->device->get_input_busy_wrapped(busy_indicator);

    if (sd->is_enabled()) {
      	sd->cancel();
      	continue;
      }

    update_needed = true;

    if (input.key == "touch_input") {
      S_Touch::get_instance()->run_callback();
      continue;
    }

    if (input.command == "back" && input.mode == "general")
      {
	exit();
      }
    else if (input.command == "next") {

      if (move_element_down())
	find_valid_program();

    } else if (input.command == "prev") {

      if (move_element_up())
	find_valid_program();

    } else if (input.command == "page_up") {

      bool changed_channel = false;

      for (int i = 0; i < conf->p_jump(); ++i) {
	if (!move_element_up())
	  break;
	else
	  changed_channel = true;
      }

      if (changed_channel)
	find_valid_program();

    } else if (input.command == "page_down") {

      bool changed_channel = false;

      for (int i = 0; i < conf->p_jump(); ++i) {
	if (!move_element_down())
	  break;
	else
	  changed_channel = true;
      }

      if (changed_channel)
	find_valid_program();

    } else if (input.command == "action") {

      print_description();

    } else if (input.command == "right") {

      find_next_program();

    } else if (input.command == "left") {

      vector<Programme>::iterator old_program_iter = program_iter;
      vector<Programme>::iterator piter;

      for (piter = sel_channel->programme.begin();
	   piter != sel_channel->programme.end(); ++piter) {
	if (piter->start > sel_time ) {
	  if (piter != sel_channel->programme.begin())
	    --piter;
	  if (piter != sel_channel->programme.begin())
	    --piter;
	  if (old_program_iter->stop == piter->stop &&
	      old_program_iter->start == piter->start)
	    if (piter != sel_channel->programme.begin())
	      --piter;

	  break;
	}
      }

      program_iter = piter;

      if (program_iter == sel_channel->programme.end()) {
	--program_iter;
	--program_iter;
      }

      sel_time = program_iter->start;

      while (sel_time < cur_time)
	cur_time -= 30*60;
    }
    else if (input.command == "options")
      {
	options();
      }
    else if (input.command == "second_action")
      {
	ExtraMenu em;

	em.add_item(ExtraMenuItem(dgettext("mms-epg", "Print description"), input_master->find_shortcut("action"),
				  boost::bind(&Epg::print_description, this)));

	em.add_item(ExtraMenuItem(dgettext("mms-epg", "Options"), input_master->find_shortcut("options"),
				  boost::bind(&Epg::options, this)));
	em.add_item(ExtraMenuItem(dgettext("mms-epg", "Return to startmenu"), input_master->find_shortcut("back"),
				  boost::bind(&Epg::exit, this)));

	foreach (ExtraMenuItem& item, global->menu_items)
	  em.add_persistent_item(item);

	em.mainloop();
      }
    else
      update_needed = !global->check_commands(input);
  }

  screen_updater->timer.del("epg");

  exit_loop = false;

  screen_updater->timer.activate("audio_fullscreen");

  return "";
}

void Epg::print()
{
  render->prepare_new_image();

  render->current.add(new PObj(themes->epg_background, 0, 0, 0, SCALE_FULL));

  PObj *back = new PObj(themes->startmenu_epg, 25, 10, 1, NOSCALING);

  render->current.add(back);

  S_Touch::get_instance()->register_area(TouchArea(rect(back->x, back->y, back->w, back->h), back->layer, boost::bind(&Epg::exit, this)));

  if (themes->show_header) {
    TObj *t = new TObj(dgettext("mms-epg", "EPG"), "Vera/28", 100, 5,
		       themes->epg_header_font1, themes->epg_header_font2,
		       themes->epg_header_font3, 1);

    render->current.add(t);

    S_Touch::get_instance()->register_area(TouchArea(rect(t->x, t->y, t->w, t->h), t->layer, global->toggle_playlist));
  }

  char time_str[30];
  strftime(time_str, 30, "%A %d-%m-%Y", localtime(&cur_time));

  int x_size = string_format::calculate_string_width(time_str, "Vera/17");

  render->current.add(new TObj(time_str, "Vera/17", (conf->p_h_res()-x_size)/2, y_start-65,
			       themes->epg_font1, themes->epg_font2,
			       themes->epg_font3, 2));

  time_t tmp_time = time(0);
  strftime(time_str, 30, "%H:%M %d-%m-%Y", localtime(&tmp_time));

  x_size = string_format::calculate_string_width(time_str, "Vera/17");

  int time_pos_from_right = 15;

  if (conf->p_busy_indicator())
    time_pos_from_right = 32;

  render->current.add(new TObj(time_str, "Vera/17", conf->p_h_res()-(x_size+time_pos_from_right), 22,
			       themes->epg_font1, themes->epg_font2,
			       themes->epg_font3, 2));

  time_t temp_time = time(0);

  double diff = 0;
  diff += (temp_time-cur_time)/(30*60);
  diff += ((temp_time-cur_time)/60-diff*30)/30.0;

  print_timelines(diff);

  if ((x_start + timeslots*30*60/scale_factor + 28) < conf->p_h_res()) {
    tmp_time = cur_time + timeslots*30*60;
    strftime(time_str, 30, "%H:%M", localtime(&tmp_time));
    render->current.add(new TObj(time_str, "Vera/17", int(x_start-28 + timeslots*30*60/scale_factor), y_start-40,
				 themes->epg_font1, themes->epg_font2, themes->epg_font3, 2));
  }

  std::ostringstream footer;
  int x, w;
  bool found_match = false;

  list<Channel>::iterator cur_chan = cur_channel;

  for (int y = 0; y < std::min(rows, int(channels.size())); ++y) {

    print_channel(y, cur_chan);

    foreach (Programme& programme, cur_chan->programme) {
      if (programme.stop > cur_time &&
	  programme.start < (cur_time + timeslots*30*60)) {

	// x = 80 + constant
	if (programme.start < cur_time) {
	  x = x_start+2;
	  w = int((programme.stop - cur_time)/scale_factor);
	} else {
	  x = int(x_start+2 + (programme.start - cur_time)/scale_factor);
	  w = int((programme.stop - programme.start)/scale_factor);
	}

	if (w < 14) // very short shows
	  w = 14;

	if (x + w > x_start+(timeslots*30*60)/scale_factor) {
	  // programs ending at the end of screen should just be cut off, others
	  // should cross the line a little to indicate that they end later.
	  w -= (x + w) - int(x_start+(timeslots*30*60)/scale_factor);
	  if (programme.stop > cur_time+timeslots*30*60)
	    w += 15;
	}

	if (cur_chan == sel_channel && programme.start <= sel_time &&
	    programme.stop > sel_time) {
	  if (!found_match) {

	    found_match = true;

	    string name = programme.title;
	    string_format::format_to_size(name, "Vera/17", w - 10, false);

	    RObj *r = new RObj(x, y*40+y_start+5, w - 5, 35,
			       themes->epg_marked1, themes->epg_marked2,
			       themes->epg_marked3, themes->epg_marked4, 3);

	    render->current.add(r);
	    render->current.add(new TObj(name, "Vera/17",
					 x+2, y*40+y_start+10, themes->epg_marked_font1,
					 themes->epg_marked_font2, themes->epg_marked_font3, 4));

	    footer.str("");
	    strftime(time_str, 30, "%H:%M", localtime(&programme.start));
	    footer << time_str << " - ";
	    strftime(time_str, 30, "%H:%M", localtime(&programme.stop));
	    footer << time_str << dgettext("mms-epg", " on ") << cur_chan->display_name;

	    string f = footer.str() + "   " + programme.title;

	    int f_size = f.size();
	    string_format::format_to_size(f, "Vera/17", conf->p_h_res()-40, false);

	    if (f.size() != f_size)
	      f += "...";

	    render->current.add(new TObj(f, "Vera/17", 15, rows*40+y_start+20, themes->epg_font1,
					 themes->epg_font2, themes->epg_font3, 4));

	    S_Touch::get_instance()->register_area(TouchArea(rect(r->x, r->y, r->w, r->h), r->layer,
							     boost::bind(&Epg::print_description, this, programme)));
	  }
	} else {

	  string name = programme.title;
	  string_format::format_to_size(name, "Vera/17", w - 10, false);

	  RObj *r = new RObj(x, y*40+y_start+5, w - 5, 35,
			     themes->epg_boxes1, themes->epg_boxes2,
			     themes->epg_boxes3, themes->epg_boxes4, 3);

 	  render->current.add(r);
	  render->current.add(new TObj(name, "Vera/17", x+2,
				       y*40+y_start+10, themes->epg_boxes_font1,
				       themes->epg_boxes_font2, themes->epg_boxes_font3, 4));

	  S_Touch::get_instance()->register_area(TouchArea(rect(r->x, r->y, r->w, r->h), r->layer,
							   boost::bind(&Epg::print_description, this, programme)));

	}
      }
    }
    ++cur_chan;
  }

  render->draw_and_release("epg");
}

void Epg::print_description(Programme p)
{
  char time_str[20];
  std::ostringstream time;

  int i = 0;
  int round = 1;
  string::iterator it = p.desc.begin();
  std::ostringstream next_word;

  while (i < p.desc.size()) {

    render->prepare_new_image();

    render->current.add(new PObj(themes->epg_background, 0, 0, 0, SCALE_FULL));

    render->current.add(new PObj(themes->startmenu_epg, 25, 10, 1, NOSCALING));

    render->current.add(new TObj("EPG", "Vera/28", 100, 10,
				 themes->epg_header_font1, themes->epg_header_font2,
				 themes->epg_header_font3, 1));

    render->current.add(new RObj(55, 90, conf->p_h_res()-2*55, conf->p_v_res()-2*93,
				 themes->general_rectangle_color1, themes->general_rectangle_color2,
				 themes->general_rectangle_color3, themes->general_rectangle_color4, 1));

    print_desc_channel(p.title);

    strftime(time_str, 20, "%H:%M", localtime(&p.start));
    time << time_str <<  " - ";
    strftime(time_str, 20, "%H:%M", localtime(&p.stop));
    time << time_str;

    render->current.add(new TObj(time.str(), "Vera/18", 75, 130,
				 themes->epg_font1, themes->epg_font2, themes->epg_font3, 2));
    time.str("");

    int start = 170;
    if (conf->p_v_res() == 405)
      start = 150;

    int y = start;

    int box_width = conf->p_h_res()-2*75;

    int lines_of_text = 12; // 800x600
    if (conf->p_h_res() == 720) {
      if (conf->p_v_res() == 540)
	lines_of_text = 10;
      else
	lines_of_text = 5;
    } else if (conf->p_h_res() == 1024) {
      lines_of_text = 18;
    }

    while (i != p.desc.size()) {
      string desc = string_format::get_line(p.desc, i, box_width, "Vera/16");

      render->current.add(new TObj(desc, "Vera/16", 75, y,
				   themes->epg_font1, themes->epg_font2,
				   themes->epg_font3, 2));
      y += 27;

      if (y == start+lines_of_text*27)
	break;
    }

    render->draw_and_release("EPG description");

    // wait for keypress
    input_master->device->get_input();

    ++round;
  }
}

void Epg::epg_print()
{
  if (conf->p_idle_time() != 0) {
    // update the idle timer
    int idle = time(0)-conf->p_last_key();
    if (idle > 60*conf->p_idle_time() && 
	(program_iter->stop > time(0) &&
	 program_iter->start < (time(0) + timeslots*30*60)))
      update_time();
  }
  print();
}

int Epg::check_epg_print()
{
  return 60*1000; // ms
}

void Epg::update_epg_datafile()
{
  if (!epg_conf->p_epg_update_script().empty()
      && file_exists(epg_conf->p_epg_update_script())
      && !update_running) {
    update_running = true;
    epg_updater->start();
  } else {
    epg_script_found = false;
    print_warning(dgettext("mms-epg", "Could not find epg update program"), "EPG");
  }
}

void EpgUpdate::run()
{
  EPGConfig *epg_conf = S_EPGConfig::get_instance();

  run::external_program(epg_conf->p_epg_update_script());

  Print pdialog(Print::SCREEN);
  pdialog.add_line(dgettext("mms-epg", "EPG TV listings file reloaded"));
  pdialog.add_line("");
  pdialog.add_line(dgettext("mms-epg", "The new data file will be used"));
  pdialog.add_line(dgettext("mms-epg", "the next time you enter EPG"));
  pdialog.print();

  // reload the db next time we enter epg
  Epg *epg = get_class<Epg>(dgettext("mms-epg", "Electronic Program Guide"));

  epg->reload_needed = false;
  epg->update_running = false;
}

int Epg::check_epg_datafile()
{
  if (!epg_conf->p_epg_data().empty()) {

    struct stat s;

    stat(epg_conf->p_epg_data().c_str(), &s);

    time_t diff = time(0) - s.st_mtime;

    int hours = diff / 60 / 60;

    if (!epg_script_found)
      return 60*1000; /* wait one minutes so that the user has time to copy
			 the script to the correct place */

    if (hours > 24 && !update_running)
      return 0; // update now
    else
      return 12*60*60*1000; // ms
  }

  return 24*60*60*1000; // just wait a long time
}

void Epg::startup_updater()
{
  ScreenUpdater *screen_updater = S_ScreenUpdater::get_instance();
  screen_updater->timer.add(TimeElement("epg_data_update", boost::bind(&Epg::check_epg_datafile, this),
					boost::bind(&Epg::update_epg_datafile, this)));

  S_Search::get_instance()->register_module(SearchModule(dgettext("mms-epg", "EPG"), boost::bind(&Epg::gen_search_list, this, _1),
							 boost::bind(&Epg::reenter, this, _1)));
}

void Epg::reenter(int db_id)
{
  if (!loaded_correctly) {
    read_db();
    loaded_correctly = true;
  }

  if (!check_tv_file()) {
    reload_needed = true;
    return;
  }

  db_mutex.enterMutex();

  SQLQuery *q = db.query("Program", ("SELECT cid, start_time, end_time FROM %t WHERE id='" + conv::itos(db_id) + "'").c_str());

  if (q && q->numberOfTuples() > 0) {

    SQLRow &row = (*q)[0];
    int start = conv::atoi(row["start_time"]);
    string tvid = row["cid"];

    int num = 0;
    for (std::list<Channel>::iterator i = channels.begin(), iend = channels.end(); i != iend; ++i, ++num)
      if (i->id == tvid) {
	sel_channel = i;
	break;
      }

    for (program_iter = sel_channel->programme.begin(),
	   program_iter_end = sel_channel->programme.end();
	 program_iter != program_iter_end; ++program_iter)
      if (program_iter->start == start)
	break;

    cur_time = sel_time = program_iter->start;

    if (channels.size() > rows) {
      sel_channel_num = rows-1;

      int start_num = num - sel_channel_num, counter = 0;
      for (std::list<Channel>::iterator i = channels.begin(), iend = channels.end(); i != iend; ++i, ++counter)
	if (counter == start_num) {
	  cur_channel = i;
	  break;
	}

    } else {
      sel_channel_num = num;
      cur_channel = channels.begin();
    }

    mainloop();
  }

  delete q;

  db_mutex.leaveMutex();
}

std::vector<std::pair<std::string, int> > Epg::gen_search_list(const std::string& search_word)
{
  std::vector<std::pair<std::string, int> > result;

  db_mutex.enterMutex();

  // check if anything has been inserted, the db doesn't get created until one
  // has entered EPG

  bool tables_found = false;

  SQLQuery *dq = db.query("sqlite_master", "SELECT name FROM %t WHERE type='table'");
  if (dq) {
    for (int i = 0; i < dq->numberOfTuples(); ++i) {
      tables_found = true;
      break;
    }
  }

  delete dq;

  if (!tables_found)
    return result;

  SQLQuery *q = db.query("Program", ("SELECT DISTINCT * FROM %t WHERE fuzzycmp('" + search_word + "', ltitle, 1)").c_str());
  if (q) {
    for (int i = 0; i < q->numberOfTuples(); ++i) {
      SQLRow &row = (*q)[i];

      string out = row["title"];

      time_t start = conv::atoi(row["start_time"]);

      char time_str[30];
      strftime(time_str, 30, "%H:%M %d-%m-%Y", localtime(&start));

      SQLQuery *q2 = db.query("Channel", ("SELECT name FROM %t WHERE tvid='" + row["cid"] + "'").c_str());
      if (q2 && q2->numberOfTuples() > 0) {
	SQLRow &c_row = (*q2)[0];
	out += string(" ") + time_str + " on " + c_row["name"];
      }
      delete q2;

      result.push_back(std::make_pair(out, conv::atoi(row["id"])));
    }
  }
  delete q;

  db_mutex.leaveMutex();

  return result;
}
