#ifndef INPUT_DEVICE_HPP
#define INPUT_DEVICE_HPP

#include "input.hpp"

typedef std::pair<std::string, int> GraphicalPair;

class BusyIndicator;

// used as superclass of input devices
class InputDevice : public ost::Thread
{
public:

  std::pair<bool, Input> generate_string(std::string& str, int& search_help_offset,
					 const boost::function<void (const std::string&)>& print_func,
					 std::list<Input> exit_inputs)
  {
    InputMaster *input_master = S_InputMaster::get_instance();

    input_master->save_map();
    input_master->set_map("search");

    Input device_input;

    std::string lowercase_str = string_format::lowercase(str);

    bool save_playlist = false;

    enable_all_input();

    generate_search_top();

    while (true) {

      bool value_found = false;

      foreach (Input& input, exit_inputs)
	if (input == device_input) {
	  value_found = true;
	  break;
	}

      if (value_found)
	break;

      print_func(str);

      device_input = get_input();

      if (device_input.key == "touch_input") {
	S_Touch::get_instance()->run_callback();
	continue;
      }

      if (device_input.command == "action") {
	save_playlist = true;
	break;
      } else {
	generate_search_input(str, lowercase_str, search_help_offset, device_input);
      }

      trigger_input(input);
    }

    disable_all_input();

    input_master->restore_map();

    return std::make_pair(save_playlist, device_input);
  }

  void register_cancel_callback(const boost::function<void (void)>& callback);
  void clear_cancel_callback();

  std::string depth;

  void set_depth(const std::string& mod)
  {
    depth = mod;
  }

  int graphical_list_size_wrapper(const boost::function<std::vector<std::pair<std::string, int> > (const std::string&)>& list,
				  const std::string& search_word)
  {
    return list(search_word).size();
  }

  bool touch_action;

  void got_touch_action()
  {
    touch_action = true;
  }

  // returns db id
  template<typename T>
  int graphical_search(const std::vector<T>& current_files,
		       const boost::function<void (const std::vector<T>&)>& print_func,
		       const boost::function<std::vector<std::pair<std::string, int> > (const std::string&)>& gen_list,
		       const boost::function<std::string (int)>& id_to_filename,
		       const boost::function<bool (void)>& is_graphical,
		       int row_size, bool& search_mode,
		       std::string& search_str, std::string& lowercase_search_str,
		       std::string& search_depth, int& offset, int& search_help_offset,
		       BusyIndicator *busy, bool show_right_selection = true)
  {
    InputMaster *input_master = S_InputMaster::get_instance();
    Config *conf = S_Config::get_instance();

    input_master->save_map();
    input_master->set_map("search");

    // reset
    search_str = "";
    lowercase_search_str = "";
    offset = 0;
    search_help_offset = -1;

    search_mode = true;

    Input device_input;

    int pos = -1;

    enable_all_input();

    search_top();

    ExtraMenu em("", true);

    if (show_right_selection) {

      em.starting_layer = 5;

      em.set_drop_down_pos(conf->p_h_res()-250+75, 75);
      em.add_item(ExtraMenuItem(gettext("current folder"), "",
				boost::bind(&InputDevice::set_depth, this,
					    gettext("current folder"))));
      em.add_item(ExtraMenuItem(gettext("all folders"), "",
				boost::bind(&InputDevice::set_depth, this,
					    gettext("all folders"))));
      em.add_exit_input(Input("left", "general"));

    }

    bool in_select_depth = false;
    depth = gettext("current folder");

    search_depth = depth;

    std::vector<T> results;

    std::string last_search_str = "1"; // different from ""

    while (device_input.command != "back") {

      if (!in_select_depth) {

	if (search_str != last_search_str) {
	  std::vector<std::pair<std::string, int> > db_results = gen_list(search_str);

	  results.clear();

	  if (depth == gettext("current folder")) {
	    foreach (const T& file, current_files) {
	      foreach (GraphicalPair& pair, db_results) {
		if (file.db_id == pair.second) {
		  results.push_back(file);
		  break;
		}
	      }
	    }
	  } else {
	    foreach (GraphicalPair& pair, db_results) {
	      std::string filename = id_to_filename(pair.second);
	      T tmp(filename);
	      tmp.name = pair.first;
	      tmp.db_id = pair.second;
	      if (isDirectory(filename))
		tmp.type = "dir";
	      else
		tmp.type = "file";
	      results.push_back(tmp);
	    }
	  }

	  if (results.size() > 0)
	    offset = 0;

	  last_search_str = search_str;
	}

	print_func(results);

	device_input = get_input();

	if (device_input.key == "touch_input") {
	  S_Touch::get_instance()->run_callback();

	  if (!touch_action)
	    continue;
	  else {
	    touch_action = false; // reset
	    pos = -1;
	    break;
	  }
	}

	if (device_input.command == "action") {

	  if (results.size() > 0) {
	    T result = vector_lookup(results, offset % results.size());

	    if (depth == gettext("current folder")) {
	      int i = 0;
	      foreach (const T& file, current_files) {
		if (file == result) {
		  pos = i;
		  break;
		} else
		  ++i;
	      }
	    } else
	      pos = result.db_id;
	    break;
	  }

	} else if ((offset == -1 || !is_graphical()) && device_input.command == "right") {

	  if (show_right_selection)
	    in_select_depth = true;

	  continue;

	} else {

	  search_graphical_input(search_str, lowercase_search_str, device_input, offset, pos,
				 search_help_offset, is_graphical(), row_size,
				 boost::bind(&InputDevice::graphical_list_size_wrapper, this, gen_list, _1));
	}

      } else if (in_select_depth) {

	em.mainloop();

	if (search_depth != depth) {
	  search_depth = depth;
	  last_search_str = search_str + "1"; // force redraw
	}

	busy_idle(busy);

	print_func(results);

	in_select_depth = false;
	continue;

      }

      trigger_input(input);
    }

    disable_all_input();

    input_master->restore_map();

    search_mode = false;

    return pos;
  }

  // recursive include work-around
  void busy_idle(BusyIndicator *indicator);

  template<typename T>
  void search(const std::vector<T>& list, int *position,
	      const boost::function<void (const std::vector<T>&)>& print_func,
	      const boost::function<bool (const T&)>& compare,
	      const boost::function<std::string (const T&)>& representation,
	      bool& search_mode, std::string& search_str, std::string& lowercase_search_str, int& offset)
  {
    InputMaster *input_master = S_InputMaster::get_instance();

    input_master->save_map();
    input_master->set_map("search");

    search_str = "";
    lowercase_search_str = "";
    offset = 0;
    search_mode = true;

    Input device_input;

    bool canceled = true;

    enable_all_input();

    search_top();

    while (device_input.command != "back") {

      print_func(gen_search(list, compare));

      device_input = get_input();

      if (device_input.key == "touch_input") {
	S_Touch::get_instance()->run_callback();
	continue;
      }

      if (device_input.command == "action") {
	canceled = false;
	break;
      } else {

	search_input(search_str, lowercase_search_str, device_input, offset,
		     boost::bind(&InputDevice::gen_searched_list<T>, this, list, _1, representation),
		     boost::bind(&InputDevice::gen_searched_list_size<T>, this, list, _1, representation));
      }
    }

    disable_all_input();

    // find position in the original list
    if (!canceled) {

      std::vector<std::string> searched = gen_searched_list(list, search_str, representation);

      if (searched.size() > 0) {
	std::string element = vector_lookup(searched, offset%searched.size());

	for (int i = 0, size = list.size(); i < size; ++i)
	  if (representation(vector_lookup(list, i)) == element) {
	    *position = i;
	    break;
	  }
      }
    }

    input_master->restore_map();

    search_mode = false;
  }

  void wait_for_input_grab()
  {
    input_done_.wait();
  }

  // helper function
  Input get_input();

  Input get_input_busy_wrapped(BusyIndicator *busy);

  // for searching
  void enable_all_input() { accept_all = true; S_ScreenUpdater::get_instance()->trigger.disable(); }
  void disable_all_input() { accept_all = false; S_ScreenUpdater::get_instance()->trigger.enable(); }

  virtual void suspend() = 0;
  virtual void wake_up() = 0;

  virtual bool init() = 0;

  virtual ~InputDevice() {};
  InputDevice();

protected:

  boost::function<void (void)> cancel_callback;

  void run_cancel_callback();

  void trigger_input(Input input);

  virtual void generate_search_top() = 0;
  virtual void generate_search_input(std::string& str, std::string& lstr, int& offset, Input input) = 0;

  virtual void search_top() = 0;

  virtual void search_input(std::string& str, std::string& lowercase_search_str, Input input, int& offset,
			    const boost::function<std::vector<std::string> (const std::string&)>& list_gen,
			    const boost::function<int (const std::string&)>& list_size) = 0;
  virtual void search_graphical_input(std::string& str, std::string& lowercase_search_str, Input input,
				      int& offset, int& pos, int& search_help_offset, bool graphical_print, int row_size,
				      const boost::function<int (const std::string&)>& list_size) = 0;

  void complete_string(std::string &search_str, std::string &lowercase_search_str,
		       const boost::function<std::vector<std::string> (const std::string&)>& list_gen)
  {
    std::vector<std::string> searched = list_gen(search_str);

    if (searched.size() != 0) {
      if (searched.size() == 1) {
	search_str = vector_lookup(searched, 0);
	lowercase_search_str = string_format::lowercase(search_str);
      } else {
	std::string gcss = string_format::greatest_common_substring(string_format::lowercase(vector_lookup(searched, 0)),
								    string_format::lowercase(vector_lookup(searched, 1)));
	for (int i = 2; i < searched.size(); ++i)
	  gcss = string_format::greatest_common_substring(gcss, string_format::lowercase(vector_lookup(searched, i)));

	search_str = gcss;
	lowercase_search_str = string_format::lowercase(search_str);
      }
    }
  }

  bool accept_all;

  Input input;

  ost::Event input_ready;
  ost::Event input_done_;

  template<typename T>
  const std::vector<T> gen_search(const std::vector<T>& list, const boost::function<bool (const T&)>& compare)
  {
    std::vector<T> newlist;

    for (int i = 0, size = list.size(); i < size; ++i)
      if (compare(vector_lookup(list, i)))
	newlist.push_back(vector_lookup(list, i));

    return newlist;
  }

  template<typename T>
  std::vector<std::string> gen_searched_list(const std::vector<T>& list, const std::string& search_word,
					     const boost::function<std::string (const T&)>& representation)
  {
    // optimization
    if (search_word.empty()) {
      std::vector<std::string> tmp;
      for (int i = 0, size = list.size(); i < size; ++i)
	tmp.push_back(representation(vector_lookup(list, i)));
      return tmp;
    }

    std::vector<std::string> newlist;

    std::string lowercase_search_str = string_format::lowercase(search_word);

    for (int i = 0, size = list.size(); i < size; ++i)
      if (string_format::lowercase(representation(vector_lookup(list, i))).substr(0, search_word.size())
	  == lowercase_search_str)
	newlist.push_back(representation(vector_lookup(list, i)));

    return newlist;
  }

  template<typename T>
  int gen_searched_list_size(const std::vector<T>& list, const std::string& search_word,
			     const boost::function<std::string (const T&)>& representation)
  {
    // optimization
    if (search_word.empty())
      return 0;

    int amount = 0;

    std::string lowercase_search_str = string_format::lowercase(search_word);

    for (int i = 0, size = list.size(); i < size; ++i)
      if (string_format::lowercase(representation(vector_lookup(list, i))).substr(0, search_word.size())
	  == lowercase_search_str)
	++amount;

    return amount;
  }

private:

  void register_keypress(Input input);

  virtual void run() = 0;
};

#endif
