#ifndef UPDATER_HPP
#define UPDATER_HPP

#include "plugins/output/image.hpp"
#include "singleton.hpp"

#include "boost.hpp"

#include <cc++/thread.h>

#include <time.h>

#include <list>
#include <string>
#include <cassert>

class TimeElement
{
public:
  typedef boost::function<void (void)> callback_func;
  typedef boost::function<int (void)> time_func;
  typedef boost::function<void (void)> cleanup_func;

  TimeElement(const std::string& n, const time_func& t, const callback_func& c)
    : name(n), active(true), time(t), callback(c), cleanup(0)
  {}

  TimeElement(const std::string& n, const time_func& t, const callback_func& c,
	      const cleanup_func& cl)
    : name(n), active(true), time(t), callback(c), cleanup(cl)
  {}

  // FIXME: add ability to add function which only runs once

  std::string name;
  bool active;
  time_func time;
  callback_func callback;
  cleanup_func cleanup;
  struct timeval next_time;
};

class TriggerElement
{
public:
  typedef boost::function<void (void)> trigger_func;
  typedef boost::function<void (void)> cleanup_func;

  TriggerElement(const std::string& n, const std::list<std::string>& trig,
		 const trigger_func& t, const cleanup_func& c)
    : name(n), triggers(trig), trigger(t), cleanup(c)
  {
    assert(t != 0);
    // clean_up can be empty if it's time controlled instead
  }

  std::string name;
  std::list<std::string> triggers;
  trigger_func trigger;
  cleanup_func cleanup;
};

class Updater : public ost::Thread
{
public:
  Updater();
  ~Updater() { terminate(); }

  void activate();
  void deactivate();
  void quit();
  bool quitting;
  void run_once(const boost::function<void (void)>& func);

  class Timer
  {
  public:
    void add(TimeElement te);
    void del(std::string name);

    // for individual elements
    void activate(const std::string& name);
    void deactivate(const std::string& name);
    void run_and_deactivate(const std::string& name);
    bool status(const std::string& name);

    struct timeval current_time;

    std::list<TimeElement> elements;

    ost::Mutex elements_mut;

    struct timeval set_timeout(const struct timeval& t, int wait);
  };

  class Trigger
  {
  public:
    Trigger();

    void add(const TriggerElement& te);
    // cleanup
    void del(const std::string& name);
    void del_all();
    // while running
    void trigger_input(const std::string& trigger);

    void disable();
    void enable();

  private:
    bool enabled;

    ost::Mutex elements_mut;

    std::list<TriggerElement> elements;
  };

  Timer timer;
  Trigger trigger;

  Overlay progressbar;
  int status_progressbar;
  int total_progressbar;
  std::string header;

private:
  struct timeval diff_timeval(const struct timeval& t1, const struct timeval& t2);

  std::list<boost::function<void (void)> > run_once_functions;

  void run();

  virtual void set_nicelevel() = 0;

  bool running;
};

class BackgroundUpdater : public Updater
{
public:
  static pthread_mutex_t singleton_mutex;

private:
  void set_nicelevel();
};

class ScreenUpdater : public Updater
{
public:
  static pthread_mutex_t singleton_mutex;

private:
  void set_nicelevel();
};

typedef Singleton<ScreenUpdater> S_ScreenUpdater;
typedef Singleton<BackgroundUpdater> S_BackgroundUpdater;

#endif
